
import ABCParse
import torch, torch_nets, neural_diffeqs


class FunctionFromDescription(ABCParse.ABCParse):
    """formatted func name from func_description"""

    def __init__(self, func_description, func_type):
        
        self.__parse__(locals(), private=['func_description'])
    
    def _simple_stripline(self, line, do_not_strip=["activation", "dropout"]):

        if not any([skip_key in line for skip_key in do_not_strip]):
            return line.replace(")", "").replace("(", "").strip(" ")
        else:
            return line.strip(" ")
    
    @property
    def raw(self):
        return self._func_description
    
    @property
    def raw_by_line(self):
        return self.raw.split("\n")
    
    @property
    def plain_lines(self):
        _plain_lines = [self._simple_stripline(line) for line in self.raw_by_line]
        return [line for line in _plain_lines if len(line) > 0]        
    
    @property
    def header(self):
        return self.plain_lines[0]
    
    # -- kwarg constructors: ------------------------------------------
    def Linear_from_line(self, line):
        split_line = line.split("=")
        in_features = int(split_line[1].split(",")[0])
        out_features = int(split_line[2].split(",")[0])
        bias = split_line[-1]

        return {"in_features": in_features, "out_features": out_features, "bias": bias}
    
    def activation_function_from_line(self, line):
        activation_line = line.split(":")[1].strip(" ").split("(")
        func, arg = activation_line[0], activation_line[1].strip(")")
        key, val = arg.split("=")
        arg = {key: float(val)}
        return {"activation": getattr(torch.nn, func)(**arg)}

    def dropout_function_from_line(self, line):
        return float(line.split("p=")[1].split(",")[0])
    
    @property
    def function_components(self):
        
        if self.func_type != "NeuralSDE":
            return ['mu']
        return ['mu', 'sigma']
    
    def _configure_network_key(self, line):
        """return mu or sigma"""
        if self.func_type == "TorchNet":
            return "mu"
        else:
            return line.split(":")[0]
        
        
    def _reconstruct_network_dict(self):
        
        # TODO: build detection of output bias
        
        NetworkComponents = {}
        for n, line in enumerate(self.plain_lines):
            _func_key = self._configure_network_key(line)
            if (not _func_key in NetworkComponents.keys()) and (_func_key in self.function_components):
                func_key = _func_key
                NetworkComponents[func_key] = {}
            elif "Sequential" in line:
                seq_key = line.split(":")[0]
                NetworkComponents[func_key][seq_key] = []
            elif "Linear" in line:
                NetworkComponents[func_key][seq_key].append(self.Linear_from_line(line))

            elif "activation" in line:
                NetworkComponents[func_key][seq_key].append(self.activation_function_from_line(line))

            elif "dropout" in line:
                NetworkComponents[func_key][seq_key].append(self.dropout_function_from_line(line))
                
        self.NetworkComponents = NetworkComponents 
        
        
    def _compose_network_kwargs(self, network_key):
                
        self.hidden, self.activation, self.dropout = [], [], []
        self._n_hidden = 0
        for key, layer in self.NetworkComponents[network_key].items():            
            if self._n_hidden == 0:
                state_size = layer[0]["in_features"]
            else:
                self.hidden.append(layer[0]["in_features"])
            if len(layer) == 3:
                self.activation.append(layer[2]['activation'])
                self.dropout.append(layer[1])
            elif len(layer) == 2:
                self.activation.append(layer[1]['activation'])
                self.dropout.append(0)
            self._n_hidden += 1
            
        if not self.func_type == "NeuralSDE":
            fmt = ""
        else:
            fmt = "{}_".format(network_key)

        self._FUNC_KWARGS["{}hidden".format(fmt)] = self.hidden
        self._FUNC_KWARGS["{}activation".format(fmt)] = self.activation
        self._FUNC_KWARGS["{}dropout".format(fmt)] = self.dropout
        
        

        if not network_key == "sigma":
            if self.func_type == "TorchNet":
                self._FUNC_KWARGS['in_features'] = self._FUNC_KWARGS['out_features'] = state_size
            else:
                self._FUNC_KWARGS['state_size'] = state_size
        
    
    def _configure_function(self):
        if not hasattr(self, "_function"):
            if self.func_type == "TorchNet":
                pkg = torch_nets
            else:
                pkg = neural_diffeqs
            self._function = getattr(pkg, self.func_type)
    
    @property
    def function(self):
        self._configure_function()
        return self._function
    
    def __call__(self):
        
        
        self._reconstruct_network_dict()
        
        self._FUNC_KWARGS = {}        
        for key in self.function_components:
            self._compose_network_kwargs(key)
            
        return self.function(**self._FUNC_KWARGS)
    
def reconstruct_function(hparams):
    func_from_description = FunctionFromDescription(hparams.func_description, hparams.func_type)
    return func_from_description()




# -- import packages: ----------------------------------------------------------
from neural_diffeqs import PotentialODE
import torch


# -- import local dependencies: ------------------------------------------------
from . import base, mix_ins


# -- set typing: ---------------------------------------------------------------
from typing import Optional, Union, List
from ... import __version__


# -- DiffEq: -------------------------------------------------------------------
class LightningODE_FixedPotential(
    mix_ins.BaseForwardMixIn,
    mix_ins.PotentialMixIn,
    base.BaseLightningDiffEq,
):
    def __init__(
        self,
        latent_dim: int = 50,
        name: Optional[str] = None,
        dt=0.1,
        mu_hidden: Union[List[int], int] = [400, 400],
        mu_activation: Union[str, List[str]] = 'LeakyReLU',
        mu_dropout: Union[float, List[float]] = 0.2,
        mu_bias: bool = True,
        mu_output_bias: bool = True,
        mu_n_augment: int = 0,
        train_lr=1e-4,
        train_optimizer=torch.optim.RMSprop,
        train_scheduler=torch.optim.lr_scheduler.StepLR,
        train_step_size=10,
        backend = "auto",
        
        adjoint=False,
        version = __version__,

        *args,
        **kwargs,
    )->None:
        """
        LightningODE_FixedPotential
        
        Parameters:
        -----------
        
        Returns:
        --------
        
        Notes:
        ------
        
        Examples:
        ---------
        """
        super().__init__()
        
        name = self._configure_name(name)

        self.save_hyperparameters()

        self._configure_torch_modules(func=PotentialODE, kwargs=locals())
        self._configure_lightning_model(kwargs = locals())

    def __repr__(self) -> str:
        return "LightningODE-FixedPotential"





# -- import packages: ----------------------------------------------------------
import pathlib
import os
from scanpy import read
import anndata
import ABCParse


# -- set typing: ---------------------------------------------------------------
from typing import Union


# -- Controller class: ---------------------------------------------------------
class PancreaticEndocrinogenesisDataset(ABCParse.ABCParse):
    """
    Pancreatic endocrinogenesis:

    *   Downloading / io taken from scvelo
    *   Data from `Bastidas-Ponce et al. (2019) <https://doi.org/10.1242/dev.173849>`__.
    *   Pancreatic epithelial and Ngn3-Venus fusion (NVF) cells during secondary transition
        with transcriptome profiles sampled from embryonic day 15.5.
    *   Endocrine cells are derived from endocrine progenitors located in the pancreatic
        epithelium. Endocrine commitment terminates in four major fates: glucagon- producing
        α-cells, insulin-producing β-cells, somatostatin-producing δ-cells and
        ghrelin-producing ε-cells.
    """

    _URL_DATADIR = "https://github.com/theislab/scvelo_notebooks/raw/master/"

    def __init__(
        self, fpath: Union[str, pathlib.Path] = "data/Pancreas/endocrinogenesis_day15.h5ad"
    ) -> None:
        """
        Parameters:
        -----------
        fpath
            Path where to save dataset and read it from.

        Returns:
        --------
        None
        """

        self.__parse__(locals(), public=[None])

    @property
    def url(self):
        return os.path.join(self._URL_DATADIR, self._fpath)

    @property
    def _PATH_EXISTS(self):
        return os.path.exists(self._fpath) 

    @property
    def _WRITE(self):
        return (not self._PATH_EXISTS) and (self._write_h5ad)

    def __call__(self, write_h5ad: bool = True) -> anndata.AnnData:
        """
        Parameters
        ----------
        write_h5ad: bool, default = True
        
        Returns
        -------
        adata: anndata.AnnData
        """
        
        self.__update__(locals(), private = [None])

        adata = read(self._fpath, backup_url=self.url, sparse=True, cache=True)
        adata.var_names_make_unique()
        if self._WRITE:
            adata.write_h5ad(self._fpath)

        return adata

# -- API-facing function: ------------------------------------------------------
def pancreas(
    fpath: Union[pathlib.Path, str] = "data/Pancreas/endocrinogenesis_day15.h5ad",
    write_h5ad: bool = True,
    *args,
    **kwargs,
) -> anndata.AnnData:

    """
    Parameters
    ----------
    fpath: Union[pathlib.Path, str], default = "data/Pancreas/endocrinogenesis_day15.h5ad"
            Path where to save dataset and/or subsequently read it from.
            
    write_h5ad: bool, default = True
        If True and the path does not exists, the file will be written to disk.

    Returns
    -------
    adata: anndata.AnnData
    """

    data = PancreaticEndocrinogenesisDataset(fpath = fpath)
    return data(write_h5ad = write_h5ad)





from ..core import utils
import ABCParse
import numpy as np
import pandas as pd


class NegativeCrossEntropy(ABCParse.ABCParse):
    def __init__(self, epsilon=1e-7):

        self.__parse__(locals(), public=[None])

    def _clip_predictions(self, pred):
        """Given a defined epsilon parameter, clip predictions"""
        return np.clip(pred, self._epsilon, 1 - self._epsilon)

    def _log_transformed_predictions(self, obs, pred_clipped):
        return obs * np.log(pred_clipped)

    def _summing(self, pred_logged):
        return -np.sum(np.sum(pred_logged, axis=1))

    def __call__(self, obs: pd.DataFrame, pred: pd.DataFrame):
        pred_clipped = self._clip_predictions(pred)
        pred_logged = self._log_transformed_predictions(obs, pred_clipped)
        return self._summing(pred_logged)

    def __repr__(self):
        return """NegativeCrossEntropy()"""



import numpy as np


class Annotator:
    def __init__(self, adata):
        self.adata = adata
        self.df = self.adata.obs.copy()

    @property
    def n_cells(self):
        return len(self.adata)

    @property
    def empty(self):
        return np.zeros(self.n_cells, dtype=bool)

    def set_true(self, idx):
        self.col = self.empty.copy()
        self.col[idx.astype(int)] = True

    @property
    def n_cells_labelled(self):
        return self.col.sum()
    
    @property
    def obs_cols(self):
        return self.df.columns.tolist()
    
    def new_col(self, key):
        """# Make a new column when one has the identical name..."""
        if key in self.obs_cols:
            cols = [col.split(".")[0] for col in self.df.columns]
            new_key = "{}.{}".format(key, cols.count(key))
        return new_key

    def message(self):
        print(
            "{} cells labelled under adata.obs['{}']".format(
                self.n_cells_labelled, self.key
            )
        )

    def __call__(self, idx, key, verbose=False):
        self.key = key
        self.set_true(idx)
        self.adata.obs[key] = self.col
        if verbose:
            self.message()


# -- API-facing function: ------------------------------------------------------------
def annotate_cells(adata, idx, key):
    """
    Given an index, create a boolean vector to be passed to model.predict(adata)
    """
    annot = Annotator(adata)
    annot(idx=idx, key=key)




from . import mix_ins, base


from neural_diffeqs import PotentialSDE
import torch_nets
import torch

from typing import Optional, Union, List

from ... import __version__

class LightningSDE_VAE_FixedPotential(
    mix_ins.PreTrainMixIn,
    mix_ins.PotentialMixIn,
    mix_ins.VAEMixIn,
    base.BaseLightningDiffEq,
):
    def __init__(
        self,
        data_dim,
        latent_dim: int = 50,
        name: Optional[str] = None,
        mu_hidden: Union[List[int], int] = [400, 400],
        mu_activation: Union[str, List[str]] = 'LeakyReLU',
        mu_dropout: Union[float, List[float]] = 0.2,
        mu_bias: bool = True,
        mu_output_bias: bool = True,
        mu_n_augment: int = 0,
        sigma_hidden: Union[List[int], int] = [400, 400],
        sigma_activation: Union[str, List[str]] = 'LeakyReLU',
        sigma_dropout: Union[float, List[float]] = 0.2,
        sigma_bias: List[bool] = True,
        sigma_output_bias: bool = True,
        sigma_n_augment: int = 0,
        sde_type='ito',
        noise_type='general',
        brownian_dim=1,
        coef_drift: float = 1.0,
        coef_diffusion: float = 1.0,
        backend = "auto",
        
        train_lr=1e-5,
        train_optimizer=torch.optim.RMSprop,
        train_scheduler=torch.optim.lr_scheduler.StepLR,
        train_step_size=10,
        
        pretrain_lr=1e-3,
        pretrain_epochs=100,
        pretrain_optimizer=torch.optim.Adam,
        pretrain_scheduler=torch.optim.lr_scheduler.StepLR,
        pretrain_step_size=200,
        
        # -- encoder parameters: -------
        encoder_n_hidden: int = 4,
        encoder_power: float = 2,
        encoder_activation: Union[str, List[str]] = 'LeakyReLU',
        encoder_dropout: Union[float, List[float]] = 0.2,
        encoder_bias: bool = True,
        encoder_output_bias: bool = True,
        
        # -- decoder parameters: -------
        decoder_n_hidden: int = 4,
        decoder_power: float = 2,
        decoder_activation: Union[str, List[str]] = 'LeakyReLU',
        decoder_dropout: Union[float, List[float]] = 0.2,
        decoder_bias: bool = True,
        decoder_output_bias: bool = True,
        
        dt=0.1,
        adjoint=False,
        version = __version__,
        
        *args,
        **kwargs,
    ):
        super().__init__()
        
        name = self._configure_name(name)

        self.save_hyperparameters()
        
        # -- torch modules: ----------------------------------------------------
        self._configure_torch_modules(func = PotentialSDE, kwargs=locals())
        self._configure_lightning_model(kwargs = locals())

    def pretrain_step(self, batch, batch_idx, stage=None):

        batch = self.process_batch(batch, batch_idx)
        X0_hat = self.Decoder(self.Encoder(batch.X0))
        recon_loss = self.reconstruction_loss(X0_hat, batch.X0).sum()
        self.log("pretrain_rl_mse", recon_loss.item())
        return recon_loss

    def training_step(self, batch, batch_idx, *args, **kwargs):
        if self.PRETRAIN:
            return self.pretrain_step(batch, batch_idx, stage="pretrain")
        return self.step(batch, batch_idx, stage="training")

    def __repr__(self):
        return "LightningSDE-VAE-FixedPotential"




import torch
from autodevice import AutoDevice
import adata_query


# -- Controller class: -----
class StateCharacterization:
    """Quantify the instantaneous drift / diffusion given a state and model."""
    def __init__(self, SDE, scalar=1):

        self.SDE = SDE
        self.scalar = scalar
    
    @property
    def t(self):
        return None
    
    def drift(self, X: torch.Tensor):
        return self.SDE.f(self.t, X).squeeze(dim=-1)
    
    def diffusion(self, X: torch.Tensor):
        return self.SDE.g(self.t, X).squeeze(dim=-1)
    
    
# -- API-facing functions: -----
def drift(adata, SDE, use_key="X_pca", key_added="X_drift", return_Tensor: bool = False):
    """Accessed as sdq.tl.drift(adata)"""
    
    X = adata_query.fetch(adata, key=use_key)
    SDE_state = StateCharacterization(SDE)
    X_drift = SDE_state.drift(X)
    adata.obsm[key_added] = X_drift.detach().cpu().numpy()
    
    if not (key_added is None):
        adata.obsm[key_added] = X_drift.detach().cpu().numpy()
    
    if return_Tensor:
        return X_drift
    
    
def diffusion(adata, SDE, use_key="X_pca", key_added="X_diffusion", return_Tensor: bool = False):
    """Accessed as sdq.tl.diffusion(adata)"""
    
    X = fetch(adata, use_key=use_key)
    SDE_state = StateCharacterization(SDE)
    X_diffusion = SDE_state.diffusion(X)
    
    if not (key_added is None):
        adata.obsm[key_added] = X_diffusion.detach().cpu().numpy()
    
    if return_Tensor:
        return X_diffusion




import tqdm
import ABCParse
from ..core import utils
from ._knn import kNN
import adata_query

NoneType = type(None)


class kNNSmoothing(ABCParse.ABCParse):
    def __init__(
        self,
        adata,
        kNN=None,
        n_iters: int = 3,
        n_neighbors: int = 20,
        use_key="X_pca",
        use_tqdm=False,
    ):

        self.__parse__(locals(), public=[None])
        self._STDEV = []

    @property
    def kNN(self):
        if isinstance(self._kNN, NoneType):
            self._kNN = kNN(
                self._adata,
                use_key=self._use_key,
                n_neighbors=self._n_neighbors,
            )
        return self._kNN

    @property
    def X_use(self):
        if not hasattr(self, "_X_use"):
            self._X_use = adata_query.fetch(
                self._adata, key=self._use_key, torch=False,
            )
        return self._X_use

    def forward(self, SCORE):
        X_nn = self.kNN.query(X_query=self.X_use)
        _SCORE = SCORE[X_nn]
        self._STDEV.append(_SCORE.std(1).mean())
        return _SCORE.mean(1)
    
    def _configure_progress_bar(self):
        if self._use_tqdm:
            return tqdm.notebook.tqdm(range(self._n_iters))
        return range(self._n_iters)
    
    @property
    def _PROGRESS_BAR(self):
        if not hasattr(self, "_progress_bar"):
            self._progress_bar = self._configure_progress_bar()
        return self._progress_bar

    @property
    def SCORE_INIT(self):
        return self._adata.obs[self._key].copy().values

    def __call__(self, key: str, add_to_adata=False):

        self._key = key

        SCORE = self.SCORE_INIT
        for i in self._PROGRESS_BAR:
            SCORE = self.forward(SCORE)

        if not add_to_adata:
            return SCORE
        else:
            self._adata.obs[f"{key}_smoothed"] = SCORE


# # -- import packages: ----------------------------------------------------------
# import anndata
# import pandas as pd
# # <<<<<<< backend
# # import ABCParse
# # import tqdm.notebook

# # =======
# # import tqdm
# # >>>>>>> pl-smooth-expr

# # -- import local dependencies: ------------------------------------------------
# from ..core import utils
# from ._grouped_expression import GroupedExpression


# # -- set typing: ---------------------------------------------------------------
# from typing import Union, List, Dict, Optional

# from typing import List


# class SmoothedExpressionSummary(ABCParse.ABCParse):
#     def __init__(
#         self,
#         keys: List[str] = ["mean", "std"],
#         storage_key: str = "t_smoothed_gex",
#         *args,
#         **kwargs,
#     ):

#         self.__parse__(locals(), public=[None])
#         self._INFO = utils.InfoMessage()

#     def _isolate_per_gene(self, key: str):
#         """
#         key:
#             "mean" or "std"
#         """
#         return pd.DataFrame(
#             {
#                 gene: values[key].tolist()
#                 for gene, values in self._adata_sim.uns[self._storage_key].items()
#             }
#         )

#     def forward(self):

#         return {
#             f"X_smoothed_gex_{key}": self._isolate_per_gene(key) for key in self._keys
#         }

#     def _add_to_adata(self, outputs):

#         for key, val in outputs.items():
#             description = key.split("_")[-1]
#             self._adata_sim.uns[key] = val
#             self._INFO(
#                 f"Smoothed {description} expression added to: `adata_sim.uns['{key}']`"
#             )

#     def __call__(self, adata_sim, *args, **kwargs):

#         self.__update__(locals(), public=[None])

#         self._add_to_adata(self.forward())


# def summarize_smoothed_expression(
#     adata_sim: anndata.AnnData,
#     keys: List[str] = ["mean", "std"],
#     storage_key: str = "t_smoothed_gex",
#     *args,
#     **kwargs,
# ) -> None:

#     """
#     Parameters
#     ----------
#     adata_sim: anndata.AnnData

#     Returns
#     -------
#     None
#     """

#     smoothed_expression_summary = SmoothedExpressionSummary(
#         keys=keys, storage_key=storage_key
#     )
#     smoothed_expression_summary(adata_sim)
    
# # -- controller class: ---------------------------------------------------------
# class SmoothedExpression(ABCParse.ABCParse):
#     def __init__(
#         self,
#         time_key: str = "t",
#         gene_id_key: str = "gene_ids",
#         use_key: str = "X_gene",
#         disable_tqdm: bool = False,
#         *args,
#         **kwargs,
#     ):

#         self.__parse__(locals(), public=[None])

#     @property
#     def _GROUPED_EXPRESSION(self):
#         if not hasattr(self, "_grouped_expr"):
#             self._grouped_expr = GroupedExpression(
#                 adata=self._adata_sim,
#                 gene_id_key=self._gene_id_key,
#                 use_key=self._use_key,
#             )
#         return self._grouped_expr

#     def _to_frame(self):
#         ...

#     def forward(self, gene_id: str):
#         grouped_vals = self._GROUPED_EXPRESSION(gene_id, groupby=self._time_key)[gene_id]
#         grouped_vals = pd.DataFrame({k: v.flatten() for k, v in grouped_vals.items()})
#         mean, std = grouped_vals.mean(0), grouped_vals.std(0)
#         return {gene_id: pd.DataFrame({"mean": mean, "std": std})}

#     @property
#     def _GENE_IDS(self):
        
#         if self._gene_id is None:
#             return self._adata_sim.uns[self._gene_id_key].tolist()
        
#         if isinstance(self._gene_id, str):
#             return [self._gene_id]
        
#         return self._gene_id

#     def _add_to_anndata(self):

        
#         uns_key = f"{self._time_key}_smoothed_gex"
        
#         if hasattr(self, "_suffix"):
#             uns_key = f"{uns_key}_{self._suffix}"

#         if not uns_key in self._adata_sim.uns:
#             self._adata_sim.uns[uns_key] = {}

#         self._adata_sim.uns[uns_key].update(self._Results)
        
#     @property
#     def _GENE_ID_PROGRESS_BAR(self):
#         if self._disable_tqdm:
#             return self._GENE_IDS
#         return tqdm.notebook.tqdm(self._GENE_IDS)

#     def __call__(
#         self,
#         adata_sim: anndata.AnnData,
#         gene_id: Optional[Union[List[str], str]] = None,
#         suffix: Optional[str] = None,
#         return_dict: bool = False,
#         *args,
#         **kwargs,
#     ):

#         self.__update__(locals(), public = [None], ignore = ['gene_id'])
#         self._gene_id = gene_id
        
#         self._Results = {}
# <<<<<<< backend
#         for gene in tqdm.notebook.tqdm(self._GENE_IDS):
# =======
#         for gene in self._GENE_ID_PROGRESS_BAR:
# >>>>>>> pl-smooth-expr
#             self._Results.update(self.forward(gene))

#         self._add_to_anndata()

#         if self._return_dict:
#             return self._Results


# class SmoothedFrameVarmHandler:
#     """
#     Operating class to mediate adding smoothed expression matrices to adata.varm
#     in the case that all genes are passed.
#     """
#     def __init__(self):
#         ...

#     def _to_frame(self, adata, key):
#         return pd.DataFrame(
#             {gene: expr[key] for gene, expr in adata.uns["t_smoothed_gex"].items()}
#         )

#     @property
#     def mean(self):
#         if not hasattr(self, "_mean"):
#             self._mean = self._to_frame(self._adata, key="mean")
#         return self._mean

#     @property
#     def std(self):
#         if not hasattr(self, "_std"):
#             self._std = self._to_frame(self._adata, key="std")
#         return self._std

#     def __call__(self, adata: anndata.AnnData, return_dfs: bool = False):

#         self._adata = adata

#         self._adata.varm["smoothed_mean"] = self.mean.T
#         self._adata.varm["smoothed_std"] = self.std.T

#         if return_dfs:
#             return self.mean, self.std

# # -- API-facing function: ------------------------------------------------------
# def smoothed_expression(
#     adata_sim: anndata.AnnData,
#     gene_id: Optional[Union[List[str], str]] = None,
#     time_key: str = "t",
#     gene_id_key: str = "gene_ids",
#     use_key: str = "X_gene",
#     return_dict: bool = False,
#     key_added_suffix: str = None,
#     *args,
#     **kwargs,
# ):
    
#     """
#     Parameters
#     ----------
#     adata_sim: anndata.AnnData
#         Simulated AnnData object.
        
#     gene_id: Optional[Union[List[str], str]], default = None
#         Gene name. If None, all genes are used. Called from adata_sim.var_names
        
#         ...
#     """
    
#     if gene_id is None:
#         gene_id = adata_sim.var_names.tolist()

#     smoothed_expression = SmoothedExpression(
#         time_key=time_key, gene_id_key=gene_id_key, use_key=use_key
#     )
# <<<<<<< backend
#     result = smoothed_expression(
#         adata_sim=adata_sim,
#         gene_id=gene_id,
#         suffix = key_added_suffix,
#         return_dict=return_dict,
#     )
    
#     storage_key = f"{time_key}_smoothed_gex"
#     if not key_added_suffix is None:
#         storage_key = "_".join([storage_key, key_added_suffix])
    
#     summarize_smoothed_expression(
#         adata_sim = adata_sim,
#         storage_key = storage_key,
#         keys = ["mean", "std"],
#     )

# =======
#     result = smoothed_expression(adata_sim = adata_sim, gene_id=gene_id, return_dict=return_dict)
    
#     if len(gene_id) == adata_sim.shape[1]:
#         varm_handler = SmoothedFrameVarmHandler()
#         varm_handler(adata_sim, return_dfs = False)
    
# >>>>>>> pl-smooth-expr
#     if return_dict:
#         return result





# -- import packages: ----------------------------------------------------------
from tqdm.notebook import tqdm
import lightning
import anndata
import pandas as pd
import torch
import glob
import os
import ABCParse
import pathlib
import autodevice

# -- import local dependencies: ------------------------------------------------
from . import configs, lightning_models, utils, callbacks
from .. import tools, __version__



import warnings

warnings.filterwarnings("ignore", ".*Consider increasing the value of the `num_workers` argument*")


# -- type setting: -------------------------------------------------------------
from typing import Union, List, Optional, Dict


import anndata
import ABCParse
from typing import Optional

class kNNMixIn(ABCParse.ABCParse, object):
    def __init__(self, *args, **kwargs):
        self.__parse__(locals())
        
    @property
    def _adata_kNN_fit(self):
        return self.adata[self.adata.obs[self._kNN_fit_subset]].copy()
        
    def _kNN_info_msg(self):
        
        self._INFO(f"Bulding Annoy kNN Graph on adata.obsm['{self._kNN_fit_subset}']")
        
    def configure_kNN(
        self,
        adata: Optional[anndata.AnnData] = None,
        kNN_key: Optional[str] = None,
        kNN_fit_subset: Optional[str] = None,
    ):
        
        """
        subset key should point to a col in adata.obs of bool vals
        """
        self.__update__(locals())
        
        self._kNN_info_msg()
        self._kNN = tools.kNN(
            adata = self._adata_kNN_fit, use_key = self._kNN_key,
        )

    @property
    def kNN(self):
        if not hasattr(self, "_kNN"):
            self.configure_kNN(
                adata = self._adata_kNN_fit,
                kNN_key = self._kNN_key,
                kNN_fit_subset = self._kNN_fit_subset,
            )
        return self._kNN

# -- model class, faces API: ---------------------------------------------------
class scDiffEq(kNNMixIn, ABCParse.ABCParse):
    def __init__(
        self,        
        
        # -- data params: -------------------------------------------------------
        adata: anndata.AnnData = None,
        latent_dim: int = 50,
        name: Optional[str] = None,
        use_key: str = "X_pca",
        obs_keys: List[str] = ["W"],
        seed: int = 0,
        backend: str = "auto",
        
        # -- kNN keys: [optional]: ----------------------------------------------
        build_kNN: Optional[bool] = False,
        kNN_key: Optional[str] = "X_pca",
        kNN_fit_subset: Optional[str] = "train",
        
        # -- pretrain params: ---------------------------------------------------
        pretrain_epochs: int = 500,
        pretrain_lr: float = 1e-3,
        pretrain_optimizer = torch.optim.Adam,
        pretrain_step_size: int = 100,
        pretrain_scheduler = torch.optim.lr_scheduler.StepLR,
        
        # -- train params: ------------------------------------------------------
        train_epochs: int = 1500,
        train_lr: float = 1e-5,
        train_optimizer = torch.optim.RMSprop,
        train_scheduler = torch.optim.lr_scheduler.StepLR,
        train_step_size: int = 10,
        train_val_split: List[float] = [0.9, 0.1],
        batch_size: int = 2000,
        
        train_key: str = "train",
        val_key: str = "val",
        test_key: str = "test",
        predict_key: str = "predict",
        
        # -- general params: ----------------------------------------------------
        logger: Optional = None,
        num_workers: int = os.cpu_count(),
        silent: bool = True,
        scale_input_counts: bool = False,
        reduce_dimensions: bool = False,
        
        fate_bias_csv_path: Optional[Union[pathlib.Path, str]] = None,
        fate_bias_multiplier: float = 1,
        viz_frequency: int = 1,
        working_dir: Union[pathlib.Path, str] = os.getcwd(),
        
        # -- time params: -------------------------------------------------------
        time_key: Optional[str] = None,
        t0_idx: Optional[pd.Index] = None,
        t_min: float = 0,
        t_max: float = 1,
        dt: float = 0.1,
        time_cluster_key: Optional[str] = None,
        t0_cluster: Optional[str] = None,
        shuffle_time_labels: bool = False,
        
        # -- DiffEq params: ----------------------------------------------------
        mu_hidden: Union[List[int], int] = [400, 400],
        mu_activation: Union[str, List[str]] = "LeakyReLU",
        mu_dropout: Union[float, List[float]] = 0.1,
        mu_bias: bool = True,
        mu_output_bias: bool = True,
        mu_n_augment: int = 0,
        
        sigma_hidden: Union[List[int], int] = [400, 400],
        sigma_activation: Union[str, List[str]] = "LeakyReLU",
        sigma_dropout: Union[float, List[float]] = 0.1,
        sigma_bias: List[bool] = True,
        sigma_output_bias: bool = True,
        sigma_n_augment: int = 0,
        
        adjoint: bool = False,
        sde_type: str = "ito",
        noise_type: str = "general",
        brownian_dim: int = 1,
        coef_drift: float = 1.0,
        coef_diffusion: float = 1.0,
        coef_prior_drift: float = 1.0,
        DiffEq_type: str = "SDE",
        potential_type: Union[None, str] = None,
        # other options: "fixed" or "prior"
        
        # -- Encoder params: ---------------------------------------------------
        encoder_n_hidden: int = 4,
        encoder_power: float = 2,
        encoder_activation: Union[str, List[str]] = "LeakyReLU",
        encoder_dropout: Union[float, List[float]] = 0.2,
        encoder_bias: bool = True,
        encoder_output_bias: bool = True,
        
        # -- Decoder params: ---------------------------------------------------
        decoder_n_hidden: int = 4,
        decoder_power: float = 2,
        decoder_activation: Union[str, List[str]] = "LeakyReLU",
        decoder_dropout: Union[float, List[float]] = 0.2,
        decoder_bias: bool = True,
        decoder_output_bias: bool = True,
        
        ckpt_path: Optional[Union[pathlib.Path, str]] = None,
        version: str = __version__,
        
        *args,
        **kwargs,
    ):
        self.__config__(locals())
        

    def _configure_dimension_reduction(self):
        self.reducer = tools.DimensionReduction(self.adata, save_path = "TESTING_SCDIFFEQ_MODEL")

        if self._scale_input_counts:
            self._INFO("Scaling input counts (for dimension reduction).")
            self.reducer.fit_scaler()
            self.adata.obsm["X_scaled_scDiffEq"] = self.reducer.X_scaled
        else:
            self.reducer.X_scaled_train = self.reducer.X_train
            self.reducer.X_scaled = self.reducer.X

        if self._reduce_dimensions:
            self.reducer.fit_pca()
            self.adata.obsm["X_pca_scDiffEq"] = self.reducer.X_pca
            self.reducer.fit_umap()
            self.adata.obsm["X_umap_scDiffEq"] = self.reducer.X_umap
            
            
    def _configure_CSVLogger(self):
        return lightning.pytorch.loggers.CSVLogger(
                save_dir=self._working_dir,
                name=self._name,
                version=None,
                prefix="",
                flush_logs_every_n_steps=1,
            )
    
    @property
    def logger(self):
        if not hasattr(self, "_LOGGER"):
            self._csv_logger = self._configure_CSVLogger()
            if hasattr(self, "_logger") and (not self._logger is None):
                    self._LOGGER = [self._csv_logger, self._logger]
            else:
                self._LOGGER = [self._csv_logger]

        return self._LOGGER

    @property
    def version(self) -> int:
        return self.logger[0].version
    
    @property
    def _metrics_path(self):
        return pathlib.Path(self.logger[0].log_dir).joinpath("metrics.csv")
    
    @property
    def metrics(self):
        return pd.read_csv(self._metrics_path)

    def _configure_trainer_generator(self):
        
        self.TrainerGenerator = configs.LightningTrainerConfiguration(self._name)
        self._PRETRAIN_CONFIG_COUNT = 0
        self._TRAIN_CONFIG_COUNT = 0

    def configure_model(self, DiffEq: Optional[lightning.LightningModule] = None):
        
        if DiffEq is None:
        
            self._LitModelConfig = configs.LightningModelConfiguration(
                data_dim=self._data_dim,
                latent_dim=self._latent_dim,
                DiffEq_type=self._DiffEq_type,
                potential_type=self._potential_type,
                fate_bias_csv_path = self._fate_bias_csv_path,
            )
            if hasattr(self, "reducer"):
                self._PARAMS['PCA'] = self.reducer.PCA
            DiffEq = self._LitModelConfig(self._PARAMS, self._ckpt_path)
            
        self.DiffEq = DiffEq
        
        self._name = self.DiffEq.hparams.name
        self._INFO(f"Using the specified parameters, {self.DiffEq} has been called.")
        self._component_loader = utils.FlexibleComponentLoader(self)
        
        lightning.seed_everything(self._seed)
        
        # -- Step 5: configure bridge to lightning logger ----------------------
        # was its own step before: now in-line here, since it
        # doesn't make sense to separate it, functionally
        self._LOGGING = utils.LoggerBridge(self.DiffEq)
        self._configure_trainer_generator()

    def configure_data(self, adata: anndata.AnnData):
        
        """Step 3. Can be called internally or externally."""
        
        self.adata = adata.copy()
        
        self._DATA_CONFIG = configs.DataConfiguration()
        self._DATA_CONFIG(scDiffEq = self)
        self._INFO(f"Input data configured.")
    
    def __config__(self, kwargs):

        """
        Run on model.__init__()

        Step 1: Parse kwargs, Set up info messaging
        Step 2: Configure data
            If adata is passed, you can do the remaining steps
            Step 3: Configure kNN [ optional ]
            Step 4: Configure model
            Step 5: Configure Logging/Trainer
            Step 6: extras (dimension reduce)
        """

        # -- Step 1: parse kwargs, set up info msg -----------------------------
        self.__parse__(kwargs, public = [None], ignore=["adata"])

        # -- TODO: eventually replace this with more sophisticated logging -----
        self._INFO = utils.InfoMessage()

        # -- Step 2: configure data ------------------------------------------
        if not kwargs['adata'] is None:
            self.configure_data(adata = kwargs['adata'])
            
            # -- Step 3: configure kNN ---------------------------------------
            if self._PARAMS["build_kNN"]:
                self._PARAMS['kNN'] = self.kNN

            # -- Step 4: configure model -------------------------------------------
            self.configure_model(DiffEq = None)

            # -- Step 6: extras: ---------------------------------------------------
#             if kwargs["reduce_dimensions"]:
#                 self._configure_dimension_reduction()

    def to(self, device):
        self.DiffEq = self.DiffEq.to(device)

    def freeze(self):
        """Freeze lightning model"""
        self.DiffEq.freeze()
        
    def load(self, ckpt_path: Union[str, pathlib.Path], freeze=True, device = autodevice.AutoDevice()):
        
        self.__update__(locals())
        
        self.DiffEq = self.DiffEq.load_from_checkpoint(self._ckpt_path)
        self.DiffEq = self.DiffEq.to(self._device)
        if freeze:
            self.freeze()

#     def _stage_log_path(self, stage):
#         log_path = glob.glob(self.DiffEqLogger.VERSIONED_MODEL_OUTDIR + f"/{stage}*")[0]
#         self._INFO(f"Access logs at: {log_path}")
        
#     @property
#     def _VERSION(self):
#         return int(os.path.basename(self.DiffEqLogger.VERSIONED_MODEL_OUTDIR).split("_")[1])
    
    def _check_disable_validation(self, trainer_kwargs):
        if self._train_val_split[1] == 0:
            trainer_kwargs.update(
                {
                    'check_val_every_n_epoch': 0,
                    'limit_val_batches': 0.0,
                    'num_sanity_val_steps': 0.0,
                    'val_check_interval': 0.0,
                },

            )
            
        return trainer_kwargs
            
    def _configure_pretrain_step(self, epochs, callbacks=[]):
        
        STAGE = "pretrain"
        self._INFO(f"Configuring fit step: {STAGE}")
        
        self.DiffEq._update_lit_diffeq_hparams(self._PARAMS)

        trainer_kwargs = utils.extract_func_kwargs(
            func=self.TrainerGenerator,
            kwargs=self._PARAMS,
            ignore = ['version', 'working_dir'],
        )
        trainer_kwargs.update(
            utils.extract_func_kwargs(
                func=lightning.Trainer,
                kwargs=self._PARAMS,
                ignore = ['version', 'working_dir'],
            )
        )
        trainer_kwargs.update(
            utils.extract_func_kwargs(
                func=lightning.Trainer,
                kwargs=locals(),
                ignore = ['version', 'working_dir'],
            )
        )
        trainer_kwargs = self._check_disable_validation(trainer_kwargs)
        
        self.pre_trainer = self.TrainerGenerator(
            max_epochs=self._pretrain_epochs,
            stage=STAGE,
            working_dir = self.DiffEqLogger._WORKING_DIR,
            version = self._VERSION,
            pretrain_version=self._PRETRAIN_CONFIG_COUNT,
            train_version=self._TRAIN_CONFIG_COUNT,
            **trainer_kwargs
        )
        self._stage_log_path(STAGE)

    def pretrain(
        self,
        epochs=None,
        pretrain_callbacks = [],
    ):
        """
        If any of the keyword arguments are passed, they will replace the previously-stated
        arguments from __init__ and re-configure the DiffEq.
        """

        self._configure_pretrain_step(epochs, pretrain_callbacks)
        self.pre_trainer.fit(self.DiffEq, self.LitDataModule)

    def _configure_train_step(self, epochs, kwargs):
                    
        STAGE = "train"
        kwargs.update(self._PARAMS)
        kwargs['callbacks'] = kwargs.pop("train_callbacks")

        self._INFO(f"Configuring fit step: {STAGE}")
        
        self.DiffEq._update_lit_diffeq_hparams(self._PARAMS)
        
        ignore = ['version', 'working_dir', 'logger']
        funcs = [self.TrainerGenerator, lightning.Trainer]
        
        trainer_kwargs = {}
        
        for func in funcs:
            trainer_kwargs.update(
                utils.extract_func_kwargs(
                    func = func, kwargs = kwargs, ignore = ignore,
                )
            )
            

#         trainer_kwargs = utils.extract_func_kwargs(
#             func=self.TrainerGenerator,
#             kwargs=self._PARAMS,
#             ignore = ignore,
#         )
#         trainer_kwargs.update(
#             utils.extract_func_kwargs(
#                 func=lightning.Trainer,
#                 kwargs=self._PARAMS,
#                 ignore = ignore,
#             )
#         )
#         trainer_kwargs.update(
#             utils.extract_func_kwargs(
#                 func=lightning.Trainer,
#                 kwargs=kwargs,
#                 ignore = ignore,
#             )
#         )
        
        self.trainer_kwargs = self._check_disable_validation(trainer_kwargs)
        
        self.trainer = self.TrainerGenerator(
            logger = self.logger,
            max_epochs=self._train_epochs,
            stage=STAGE,
            working_dir = self._working_dir,
            version = self.version,
            pretrain_version=self._PRETRAIN_CONFIG_COUNT,
            train_version=self._TRAIN_CONFIG_COUNT,
#             callbacks = train_callbacks,
            **self.trainer_kwargs,
        )
        
        if hasattr(self, "_csv_logger"):
            DIR = self._csv_logger.log_dir
            self._INFO(f"Logging locally to: {DIR}")
        
        
        self._TRAIN_CONFIG_COUNT += 1

#         self._stage_log_path(STAGE)

    def train(
        self,
        epochs=500,
        train_callbacks=[],
        ckpt_frequency: int = 25,
        save_last_ckpt: bool = True,
        keep_ckpts: int = -1,
        monitor=None,
        accelerator=None,
        log_every_n_steps=1,
        reload_dataloaders_every_n_epochs=1,
        devices=None,
        deterministic=False,
        **kwargs,
    ):
                
        self.DiffEq._update_lit_diffeq_hparams(self._PARAMS)
        self._configure_train_step(epochs, locals())
        self.trainer.fit(self.DiffEq, self.LitDataModule)
        
    def fit(
        self,
        train_epochs=200,
        pretrain_epochs=500,
        train_lr = None,
        pretrain_callbacks: List = [],
        train_callbacks: List = [],
        ckpt_frequency: int = 25,
        save_last_ckpt: bool = True,
        keep_ckpts: int = -1,
        monitor=None,
        accelerator=None,
        log_every_n_steps=1,
        reload_dataloaders_every_n_epochs=1,
        devices=1,
        deterministic=False,
        **kwargs,
    ):
                
        if pretrain_epochs > 0 and (not self._LitModelConfig.use_vae):
            pretrain_epochs = 0
            
            
        self.__update__(locals())

        if pretrain_epochs > 0:
            self.pretrain(
                epochs=pretrain_epochs, **utils.extract_func_kwargs(self.pretrain, locals()))
        if train_epochs > 0:
            self.train(
                epochs=train_epochs, **utils.extract_func_kwargs(self.train, locals())
            )
            
        # TO-DO: eventually replace how this works...
        self._PRETRAIN_CONFIG_COUNT += 1
#         self._TRAIN_CONFIG_COUNT += 1
        
    @property
    def tracker(self):
        return callbacks.ModelTracker(version=self.version)


    def cell_potential(
        self,
        use_key: str = 'X_pca',
        raw_key_added: str = '_psi',
        norm_key_added: str = 'psi',
        device: Union[str, torch.device] = torch.device('cuda:0'),
        seed: int = 0,
        normalize: bool = True,
        return_raw_array: bool = False,
        q: float = 0.05,
        knn_smoothing_iters: int = 5,
        use_tqdm: bool = True,
    ):

        tools.cell_potential(
            adata = self.adata,
            model = self,
            use_key = use_key,
            raw_key_added = raw_key_added,
            norm_key_added = norm_key_added,
            device = device,
            seed = seed,
            normalize = normalize,
            return_raw_array = return_raw_array,
            q = q,
            knn_smoothing_iters = knn_smoothing_iters,
            use_tqdm = use_tqdm,
        )

            

# if you passed adata, configure the rest of the model here:
        # -- 1a. parse adata: --------------------------------------------------
#             self._DATA_CONFIG = configs.DataConfiguration()
#             self._DATA_CONFIG(
#                 adata = self.adata, scDiffEq = self, scDiffEq_kwargs = self._PARAMS,
#             )
        
#     # -- could cordon off into a kNNMixIn: ----
#     def _configure_kNN_graph(self):      
#         train_adata = self.adata[self.adata.obs[self._train_key]].copy()
#         self._INFO(f"Bulding Annoy kNN Graph on adata.obsm['{self._kNN_key}']")        
#         self._kNN_graph = tools.kNN(adata = train_adata, use_key = self._kNN_key)
        
#     @proprety
#     def kNN(self):
#         if not hasattr(self, "_kNN_graph"):
#             self._configure_kNN_graph()
#         return self._kNN_graph
#     # -----------------------------------------



#     @property
#     def loss(self):
#         utils.display_tracked_loss(self.logger)
        
# -- ADD AS MIXINS: ----
#     def load_DiffEq_from_ckpt(self, ckpt_path):
        
#         self._component_loader.load_DiffEq_state(ckpt_path)
#         self._PARAMS[
#             "diffeq_ckpt_path"
#         ] = self._diffeq_ckpt_path = self._component_loader._diffeq_ckpt_path
#         self.DiffEq._update_lit_diffeq_hparams(self._PARAMS)
        
                
#     def load_encoder_from_ckpt(self, ckpt_path):
        
#         self._component_loader.load_encoder_state(ckpt_path)
#         self._PARAMS[
#             "encoder_ckpt_path"
#         ] = self._encoder_ckpt_path = self._component_loader._encoder_ckpt_path
#         self.DiffEq._update_lit_diffeq_hparams(self._PARAMS)
        
        
#     def load_decoder_from_ckpt(self, ckpt_path):
        
#         self._component_loader.load_decoder_state(ckpt_path)
#         self._PARAMS[
#             "decoder_ckpt_path"
#         ] = self._decoder_ckpt_path = self._component_loader._decoder_ckpt_path
#         self.DiffEq._update_lit_diffeq_hparams(self._PARAMS)
        
#     def load_VAE_from_ckpt(self, ckpt_path):
#         # TODO: add ability to freeze these once loaded
        
#         self.load_encoder_from_ckpt(ckpt_path)
#         self.load_decoder_from_ckpt(ckpt_path)




# -- import packages: ---------------------------------------------------------
import ABCParse
import torch
import numpy as np


# -- set typing: --------------------------------------------------------------
from typing import Union, List


class BatchRepresentation:
    def __init__(self, batch):

        self.header = "🍪 LitDataBatch"
        self.X = f"- X: {list(batch.X.shape)}"
        self.W = f"- W: {list(batch.W.shape)}"
        self.W_hat = f"- W_hat: {list(batch.W_hat.shape)}"
        if not batch.F_idx is None:
            self.F_idx = f"- F_idx: {list(batch.F_idx.shape)}"
        else:
            self.F_idx = ""

    def __call__(self):
        return "\n".join([self.header, self.X, self.W, self.W_hat, self.F_idx])


class BatchProcessor(ABCParse.ABCParse):
    """
    Batch Items:
    ------------
    0. t - time. always required
    1. X - data. tensor of size: [batch_size, len(t), n_dim]. always required.
    2. W - weight. tensor of size: [batch_size, len(t), 1]
    3. F_idx - fate_idx. tensor of size: [batch_Size, len(t), 1]
    """

    def __init__(self, batch: List[torch.Tensor], batch_idx: int) -> None:

        self.__parse__(locals(), private=["batch"])

    def _sum_normalize(
        self, X: Union[torch.Tensor, np.ndarray], sample_axis=1
    ) -> Union[torch.Tensor, np.ndarray]:
        """
        Parameters:
        -----------
        X
            torch.Tensor or nump.array

        sample_axis
            default: 1
            type: int

        Returns:
        --------
        return X_
        """
        return X / X.sum(sample_axis)[:, None]

    @property
    def device(self):
        return self._batch[0].device

    @property
    def n_batch_items(self) -> int:
        return len(self._batch)

    @property
    def batch_size(self) -> int:
        return self._batch[0].shape[0]

    @property
    def t(self) -> torch.Tensor:
        return self._batch[0].unique()

    @property
    def X(self) -> torch.Tensor:
        return self._batch[1].transpose(1, 0).contiguous()

    @property
    def X0(self) -> torch.Tensor:
        return self.X[0]

    @property
    def W_hat(self) -> torch.Tensor:
        if self.n_batch_items >= 3:
            W = self._batch[2].transpose(1, 0)
            
            for i in range(1, len(self.t)):
                W[i] = W[0] * torch.exp(self.t[i] - self.t[0])
            W[0] = torch.ones_like(W[0])
            
            return self._sum_normalize(W, sample_axis=1).contiguous()

        return self._sum_normalize(
            torch.ones([len(self.t), self.batch_size, 1], device=self.device)
        ).contiguous()

    @property
    def W(self) -> torch.Tensor:
        return self._sum_normalize(torch.ones_like(self.W_hat)).contiguous()

    @property
    def F_idx(self):
        if self.n_batch_items >= 4:
            return (
                self._batch[3]
                .transpose(1, 0)[0]
                .detach()
                .cpu()
                .numpy()
                .flatten()
                .astype(int)
                .astype(str)
            )

    def __repr__(self) -> str:
        return BatchRepresentation(self)()




__author__ = "Michael E. Vinyard"
__email__ = "mvinyard.ai@gmai.com"


# -- import models accessed as sdq.models.<MODEL>: ---------------------------------------
from ._scdiffeq import scDiffEq

from . import lightning_models
from . import utils
from . import configs
from . import callbacks



class DriftPriorMixIn(object):
    logqp = True
    def __init__(self, *args, **kwargs):
        super().__init__()

    def forward(self, Z0, t, **kwargs):
        """Forward step: (0) integrate in latent space"""
        
        return self.integrate(
            Z0=Z0,
            t=t,
            dt=self.hparams["dt"],
            logqp=self.logqp,
            **kwargs,
        )
            
    def log_computed_loss(self, sinkhorn_loss, t, kl_div_loss, stage):
        
        sinkhorn_loss = self.log_sinkhorn_divergence(
            sinkhorn_loss=sinkhorn_loss,
            t=t,
            stage=stage,
        ).sum()
        self.log(f"kl_div_{stage}", kl_div_loss.sum())

        return sinkhorn_loss + kl_div_loss.sum()

    def step(self, batch, batch_idx, stage=None):

        batch = self.process_batch(batch, batch_idx)
        X_hat, kl_div_loss = self.forward(batch.X0, batch.t)
        sinkhorn_loss = self.compute_sinkhorn_divergence(
            batch.X, X_hat, batch.W, batch.W_hat
        )
        return self.log_computed_loss(
            sinkhorn_loss, t=batch.t, kl_div_loss=kl_div_loss, stage=stage
        )

class DriftPriorVAEMixIn(object):
    logqp = True
    def __init__(self, *args, **kwargs):
        super().__init__()

    def forward(self, X0, t, logqp=True, **kwargs):
        """Forward step: (0) integrate in latent space"""
        
        Z_hat, kl_div_loss = self.integrate(
            Z0=self.Encoder(X0),
            t=t,
            dt=self.hparams["dt"],
            logqp=logqp,
            **kwargs,
        )
        
        return self.Decoder(Z_hat), kl_div_loss
    
    def log_computed_loss(self, sinkhorn_loss, t, kl_div_loss, stage):
        
        sinkhorn_loss = self.log_sinkhorn_divergence(
            sinkhorn_loss=sinkhorn_loss,
            t=t,
            stage=stage,
        ).sum()
        self.log(f"kl_div_{stage}", kl_div_loss.sum())

        return sinkhorn_loss + kl_div_loss.sum()

    def step(self, batch, batch_idx, stage=None):
        
        train_optim = self.optimizers()[1]
        train_optim.zero_grad()
        
        batch = self.process_batch(batch, batch_idx)
        X_hat, kl_div_loss = self.forward(batch.X0, batch.t)
        sinkhorn_loss = self.compute_sinkhorn_divergence(
            batch.X, X_hat, batch.W, batch.W_hat
        )
        loss = self.log_computed_loss(
            sinkhorn_loss, t=batch.t, kl_div_loss=kl_div_loss, stage=stage
        )
        self.manual_backward(loss)
        train_optim.step()




# -- import packages: ----------------------------------------------------------
from neural_diffeqs import NeuralODE
import torch


# -- import local dependencies: ------------------------------------------------
from . import base, mix_ins


from typing import Optional, Union, List
from ... import __version__


# -- lightning model: ----------------------------------------------------------
class LightningODE(
    mix_ins.BaseForwardMixIn,
    base.BaseLightningDiffEq,
):
    def __init__(
        self,
        # -- ode params: -------------------------------------------------------
        latent_dim: int = 50,
        name: Optional[str] = None,
        mu_hidden: Union[List[int], int] = [2000, 2000],
        mu_activation: Union[str, List[str]] = 'LeakyReLU',
        mu_dropout: Union[float, List[float]] = 0.2,
        mu_bias: bool = True,
        mu_output_bias: bool = True,
        mu_n_augment: int = 0,
        sde_type='ito',
        noise_type='general',
        backend = "auto",
        
        # -- general params: ---------------------------------------------------
        train_lr: float = 1e-4,
        train_optimizer = torch.optim.RMSprop,
        train_scheduler = torch.optim.lr_scheduler.StepLR,
        train_step_size: int = 10,
        dt: float = 0.1,
        adjoint=False,
        version = __version__,
        *args,
        **kwargs,
    ):
        super().__init__()
        
        name = self._configure_name(name)

        self.save_hyperparameters()
        
        # -- torch modules: ----------------------------------------------------
        self._configure_torch_modules(func=NeuralODE, kwargs=locals())
        self._configure_lightning_model(kwargs = locals())

    def __repr__(self):
        return "LightningODE"




import pickle


class PickleIO:
    def __init__(self):
        pass

    def read(self, path, mode="rb"):
        return pickle.load(self.__path__(path, mode))

    def write(self, obj, path, mode="wb", protocol=pickle.HIGHEST_PROTOCOL):
        """If writing for use in colab, use protocol=4"""

        pickle.dump(obj=obj, file=self.__path__(path, mode), protocol=protocol)

    def __path__(self, path, mode):
        return open(path, mode)
    
    
def read_pickle(path, mode="rb"):

    pickle_io = PickleIO()
    return pickle_io.read(path, mode=mode)


def write_pickle(obj, path, mode="wb", protocol=pickle.HIGHEST_PROTOCOL):

    pickle_io = PickleIO()
    return pickle_io.write(obj=obj, path=path, mode=mode, protocol=protocol)




__module_name__ = "__init__.py"
__doc__ = """I/O Functions: h5ad interactions."""
__author__ = ", ".join(["Michael E. Vinyard",])
__email__ = ", ".join(["vinyard@g.harvard.edu",])


# import packages: -----------------------------------------------------------------------
import anndata


# API-facing function: -------------------------------------------------------------------
def read_h5ad(h5ad_path: str, silent: bool = False, annotate_path: bool = True) -> anndata.AnnData:

    """
    Read an AnnData object from a path to a .h5ad file.

    Parameters:
    -----------
    h5ad_path: str
        path to .h5ad file.

    silent: bool, default = False
        If True, the anndata object is not printed.
        
    annotate_path: bool, default = True
        If True, adds the h5ad_path to the read adata object in adata.uns['h5ad_path'].

    Returns:
    --------
    adata
        type: anndata._core.anndata.AnnData
        

    Notes:
    ------
    (1) Documentation for AnnData: https://anndata.readthedocs.io/en/stable/

    """

    adata = anndata.read_h5ad(h5ad_path)
    adata.uns['h5ad_path'] = h5ad_path

    if not silent:
        print(adata)

    return adata


import lightning
import torch

from typing import Any, Dict, Union

class ModelLogging(lightning.Callback):
    def __init__(self, *args, **kwargs):
        
        self._CURRENT_EPOCH = None
            
    def on_train_epoch_end(self, trainer, pl_module):
        
        epoch_train_loss = torch.hstack(self._EPOCH_TRAIN_LOSS).mean()
        pl_module.log("epoch_train_loss", epoch_train_loss)
        
    def on_validation_epoch_end(self, trainer, pl_module):
        
        epoch_validation_loss = torch.hstack(self._EPOCH_VALIDATION_LOSS).mean()
        pl_module.log("epoch_validation_loss", epoch_validation_loss)
        
    def log_total_epochs(self, pl_module):
        """Train model N times --> N"""
        pl_module.log("total_epochs", pl_module.COMPLETED_EPOCHS)

    def log_lr(self, pl_module):

        if not isinstance(pl_module.optimizers(), list):
            lr = pl_module.optimizers().optimizer.state_dict()["param_groups"][0]["lr"]
            pl_module.log("opt_param_group_lr", lr)
        else:
            for i, opt in enumerate(pl_module.optimizers()):
                for j, pg in enumerate(opt.optimizer.state_dict()["param_groups"]):
                    pl_module.log(f"opt_{i}_param_group_{j}_lr", pg["lr"])

    def log_sinkhorn_divergence(self, pl_module, t, stage: str):

        sinkhorn_loss = pl_module.sinkhorn_loss
        
        for i in range(len(t)):
            _t = round(t[i].item(), 3)
            pl_module.log(f"sinkhorn_{_t}_{stage}", sinkhorn_loss[i])

        return sinkhorn_loss.sum()

    def _gather_current_epoch_loss(self, pl_module, sinkhorn_total, stage):
        
        if pl_module.COMPLETED_EPOCHS != self._CURRENT_EPOCH:
            self._EPOCH_TRAIN_LOSS = []
            self._EPOCH_VALIDATION_LOSS = []
            self._CURRENT_EPOCH = pl_module.COMPLETED_EPOCHS
            
        else:
            if stage == "training":
                self._EPOCH_TRAIN_LOSS.append(sinkhorn_total.detach().cpu())
            elif stage == "validation":
                self._EPOCH_VALIDATION_LOSS.append(sinkhorn_total.detach().cpu())

    def on_train_batch_end(
        self, 
        trainer: 'pl.Trainer',
        pl_module: 'pl.LightningModule',
        outputs: Union[torch.Tensor, Dict[str, Any]],
        batch: Any,
        batch_idx: int,
    ):
        
        batch = pl_module.process_batch(batch, batch_idx)
        
        self.log_total_epochs(pl_module)
        self.log_lr(pl_module)
        sinkhorn_total = self.log_sinkhorn_divergence(
            pl_module = pl_module, t = batch.t, stage = "training",
        )
        self._gather_current_epoch_loss(
            pl_module = pl_module, sinkhorn_total = sinkhorn_total, stage = "training",
        )

    def on_validation_batch_end(
        self, 
        trainer: 'pl.Trainer',
        pl_module: 'pl.LightningModule',
        outputs: Union[torch.Tensor, Dict[str, Any]],
        batch: Any,
        batch_idx: int,
    ):
        
        batch = pl_module.process_batch(batch, batch_idx)
        
        self.log_total_epochs(pl_module)
        self.log_lr(pl_module)
        sinkhorn_total = self.log_sinkhorn_divergence(
            pl_module = pl_module, t = batch.t, stage = "validation",
        )
        self._gather_current_epoch_loss(
            pl_module = pl_module, sinkhorn_total = sinkhorn_total, stage = "validation",
        )



import abc
import numpy as np
import torch
from typing import Union


class FlexibleCall(abc.ABC):
    def __init__(self, *args, **kwargs):

        """"""

    @property
    def dtype(self):
        return type(self.input)

    @property
    def flexible_forward(self):
        if isinstance(self.input, np.ndarray):
            return self._numpy
        if isinstance(self.input, torch.Tensor):
            return self._torch

    @abc.abstractmethod
    def _numpy(self, input: np.ndarray, *args, **kwargs):
        ...

    @abc.abstractmethod
    def _torch(self, input: torch.Tensor, *args, **kwargs):
        ...

    def __call__(
        self, input: Union[np.ndarray, torch.Tensor], *args, **kwargs
    ) -> Union[np.ndarray, torch.Tensor]:
        self.input = input
        return self.flexible_forward(self.input, *args, **kwargs)


class L1Norm(FlexibleCall):
    """Manhattan Distance"""

    def __init__(self, *args, **kwargs):
        super().__init__()

    def _numpy(self, input: np.ndarray, axis=-1, **kwargs):
        return np.linalg.norm(input, ord=1, axis=axis, **kwargs)

    def _torch(self, input: torch.Tensor, dim=-1, **kwargs):
        return input.norm(p=1, dim=dim, **kwargs)

    def __repr__(self):
        return "L1 Norm: Manhattan Distance"


class L2Norm(FlexibleCall):
    """Euclidean Distance"""

    def __init__(self):
        super().__init__()

    def _numpy(self, input: np.ndarray, axis=-1, **kwargs):
        return np.linalg.norm(input, ord=2, axis=axis, **kwargs)

    def _torch(self, input: torch.Tensor, dim=-1):
        return input.norm(p=2, dim=dim)

    def __repr__(self):
        return "L2 Norm: Euclidean Distance"



import os, glob, numpy as np
import ABCParse


class Versions(ABCParse.ABCParse):
    def __init__(
        self, base_dir=".", base_path="scDiffEq_model/lightning_logs/version_{}/hparams.yaml"
    ):

        self.__parse__(locals(), private=["base_dir", "base_path"])

    def __basedir__(self, path):
        return os.path.basename(os.path.dirname(path))

    @property
    def base_path(self):
        return os.path.join(self._base_dir, self._base_path)

    @property
    def available(self):
        return [
            self.__basedir__(version)
            for version in glob.glob(self.base_path.format("*"))
        ]

    @property
    def indices(self):
        return [int(version.split("_")[1]) for version in self.available]

def configure_version(
    version=None,
    base_dir=".",
    base_path="scDiffEq_model/lightning_logs/version_{}/hparams.yaml",
):
    """returns version, yaml_base_path"""

    _versions = Versions(base_dir=base_dir, base_path=base_path)
    base_path = _versions.base_path
    available = _versions.available
    if not version:
        version = _versions.indices[np.argmax(_versions.indices)]
    return {"version": version, "base_path": base_path, "available": available}





# -- import packages: ----------------------------------------------------------
import adata_query
import sklearn
import pickle
import umap
import os
import ABCParse


# -- import local dependencies: ------------------------------------------------
from ..core import utils
# from ._x_use import fetch_formatted_data


# -- type setting: -------------------------------------------------------------
NoneType = type(None)


# -- controller class: ---------------------------------------------------------
class DimensionReduction(ABCParse.ABCParse):
    _SCALER_FIT, _PCA_FIT, _UMAP_FIT = [False] * 3

    def __init__(
        self,
        adata,
        use_key: str = "X",
        n_pcs: int = 50,
        n_umap_components: int = 2,
        train_key="train",
        test_key="test",
        write=True,
        save_path="./reducer_models",
        use_saved=True,
        pkl_protocol=4,
        scaler_kwargs={},
        pca_kwargs={},
        umap_kwargs={},
        *args,
        **kwargs,
    ):

        self._configure(locals())
        
    def _configure(self, kwargs):

        """
        
        * If train_key is not in adata, set all cells to train.
        """
        self.__parse__(kwargs, public=[None])
        
        if not self._train_key in self._adata.obs.columns:
            self._adata.obs[self._train_key] = True
            
        self.df = self._adata.obs.copy()
        self._INFO = utils.InfoMessage()

        if not os.path.exists(self._save_path):
            self._INFO(f"mkdir: {self._save_path}")
            os.mkdir(self._save_path)

    def _configure_kwargs(self, func, kwarg_subset_key, specific_kwarg: dict = None):
        """"""
        PARAMS = self._PARAMS
        PARAMS.update(self._PARAMS[kwarg_subset_key])
        if not isinstance(specific_kwarg, NoneType):
            PARAMS.update(specific_kwarg)
        return ABCParse.function_kwargs(func=func, kwargs=PARAMS)

    @property
    def _SCALER_MODULE(self):
        return sklearn.preprocessing.StandardScaler

    @property
    def _PCA_MODULE(self):
        return sklearn.decomposition.PCA

    @property
    def _UMAP_MODULE(self):
        return umap.UMAP

    @property
    def _SCALER_KWARGS(self):
        return self._configure_kwargs(
            func=self._SCALER_MODULE,
            kwarg_subset_key="scaler_kwargs",
        )

    @property
    def _PCA_KWARGS(self):
        return self._configure_kwargs(
            func=self._PCA_MODULE,
            kwarg_subset_key="pca_kwargs",
            specific_kwarg={"n_components": self._n_pcs},
        )

    @property
    def _UMAP_KWARGS(self):
        return self._configure_kwargs(
            func=self._UMAP_MODULE,
            kwarg_subset_key="umap_kwargs",
            specific_kwarg={"n_components": self._n_umap_components},
        )

    @property
    def train_idx(self):
        if not hasattr(self, "_train_idx"):
            self._train_idx = self._adata[self.df[self._train_key]].obs.index
        return self._train_idx

    @property
    def test_idx(self):
        if not hasattr(self, "_test_idx"):
            self._test_idx = self._adata[self.df[self._test_key]].obs.index
        return self._test_idx

    @property
    def unique_test_idx(self):
        if not hasattr(self, "_unique_test_idx"):
            train, test = self.train_idx, self.test_idx
            self._unique_test_idx = test[[idx not in train for idx in test]]
        return self._unique_test_idx

    @property
    def X_train(self):
        if not hasattr(self, "_X_train"):
            self._X_train = adata_query.fetch(
                adata=self._adata[self.train_idx],
                key=self._use_key,
                torch=False,
            )

        return self._X_train

    @property
    def X(self):
        if not hasattr(self, "_X"):
            self._X = adata_query.fetch(
                adata=self._adata,
                key=self._use_key,
                torch=False,
            )

        return self._X

    @property
    def _N_CELLS_TRAIN(self):
        return self.X_train.shape[0]

    @property
    def _N_CELLS(self):
        return self.X.shape[0]

    @property
    def _FIT_USING_ALL_CELLS(self):
        return self._N_CELLS_TRAIN == self._N_CELLS

    # -- paths to saved models: ------------------------------------------------
    @property
    def _PATH_SCALER(self):
        return os.path.join(self._save_path, "scaler_model.pkl")

    @property
    def _PATH_PCA(self):
        return os.path.join(self._save_path, "pca_model.pkl")

    @property
    def _PATH_UMAP(self):
        return os.path.join(self._save_path, "umap_model.pkl")

    def _msg(self, PATH):

        if not os.path.exists(PATH):
            self._INFO(f"PATH: {PATH} DOES NOT EXIST.")
        else:
            if self._use_saved:
                use = "Will use this model."
            else:
                use = "Not using."
            self._INFO(f"PATH: {PATH} DOES EXIST. {use}")

    def _OK_TO_LOAD(self, PATH):
        return os.path.exists(PATH) and (self._use_saved)

    def _LOAD(self, PATH, key):
        self._INFO(f"Loading {key} from: {PATH}")
        return pickle.load(open(PATH, "rb"))

    def _SETUP_MODULE(self, key):
        MODULE = getattr(self, f"_{key}_MODULE")
        KWARGS = getattr(self, f"_{key}_KWARGS")
        return MODULE(**KWARGS)

    def _CONFIGURE_MODEL(self, key):
        self._INFO(f"Configuring: {key} model")
        """try (1) loading from file if it exists then (2) building it from scratch"""

        PATH = getattr(self, f"_PATH_{key}")
        self._msg(PATH)
        if self._OK_TO_LOAD(PATH):
            setattr(self, f"_{key}_FIT", True)
            return self._LOAD(PATH, key)

        return self._SETUP_MODULE(key)
    
    # -- models: ---------------------------------------------------------------
    @property
    def SCALER(self):
        if not hasattr(self, "_SCALER_MODEL"):
            self._SCALER_MODEL = self._CONFIGURE_MODEL("SCALER")
        return self._SCALER_MODEL

    @property
    def PCA(self):
        if not hasattr(self, "_PCA_MODEL"):
            self._PCA_MODEL = self._CONFIGURE_MODEL("PCA")
        return self._PCA_MODEL

    @property
    def UMAP(self):
        if not hasattr(self, "_UMAP_MODEL"):
            self._UMAP_MODEL = self._CONFIGURE_MODEL("UMAP")
        return self._UMAP_MODEL

    def _write_model(self, MODEL, key):
        PATH = getattr(self, f"_PATH_{key}")
        self._INFO(f"Saving {key} model to: {PATH}")
        pickle.dump(MODEL, open(PATH, "wb"))

    def _fit(
        self,
        X_input,
        X_input_train,
        key,
    ):

        MODEL = getattr(self, key)
        FIT = getattr(self, f"_{key}_FIT")

        if FIT:
            self._INFO(f"Using previously fit {key} model")
            if (
                (key == "UMAP")
                and self._FIT_USING_ALL_CELLS
                and (MODEL.embedding_.shape[0] == self._N_CELLS)
            ):
                self._INFO("Using cached embedding from UMAP model")
                X_transformed_train = MODEL.embedding_
            else:
                X_transformed_train = MODEL.transform(X_input_train)
            if self._FIT_USING_ALL_CELLS:
                X_transformed = X_transformed_train
            else:
                X_transformed = MODEL.transform(X_input)
        else:
            self._INFO(f"Fitting {key} model")

            X_transformed_train = MODEL.fit_transform(X_input_train)

            if self._FIT_USING_ALL_CELLS:
                X_transformed = X_transformed_train
            else:
                X_transformed = MODEL.transform(X_input)

            self._write_model(MODEL, key)

        return MODEL, X_transformed, X_transformed_train

    # -- fit functions: --------------------------------------------------------
    def fit_scaler(self):

        self._SCALER_MODEL, self.X_scaled, self.X_scaled_train = self._fit(
            X_input=self.X,
            X_input_train=self.X_train,
            key="SCALER",
        )
        self._adata.layers["X_scaled"] = self.X_scaled

    def fit_pca(self):

        self._PCA_MODEL, self.X_pca, self.X_pca_train = self._fit(
            X_input=self.X_scaled,
            X_input_train=self.X_scaled_train,
            key="PCA",
        )
        self._adata.obsm["X_pca"] = self.X_pca

    def fit_umap(self):

        self._UMAP_MODEL, self.X_umap, self.X_umap_train = self._fit(
            X_input=self.X_pca,
            X_input_train=self.X_pca_train,
            key="UMAP",
        )
        self._adata.obsm["X_umap"] = self.X_umap





import scdiffeq_plots as sdq_pl
import numpy as np
import ABCParse

from ..core import utils
        
class FinalStatePerSimulation(ABCParse.ABCParse):
    """"""

    def __init__(
        self,
        adata,
        obs_key="Cell type annotation",
        N_sim=2000,
        time_key="t",
        sim_key="simulation",
    ):
        self.__parse__(locals(), public=["adata"])
        self._configure_simulation_key()

    @property
    def _MAX_TIME(self):
        return self.adata.obs[self._time_key].max()

    @property
    def _N_TIMEPOINTS(self):
        return self.adata.obs[self._time_key].nunique()

    @property
    def _SIM_COL(self):
        if not hasattr(self, "_sim_col"):
            self._sim_col = np.tile(np.arange(self._N_sim), self._N_TIMEPOINTS)
        return self._sim_col

    @property
    def _SIMULATION_GROUPED(self):
        if not hasattr(self, "_sim_grouped"):
            self._sim_grouped = self.adata.obs.groupby(self._sim_key)
        return self._sim_grouped

    def _configure_simulation_key(self):
        if not self._sim_key in self.adata.obs:
            self.adata.obs[self._sim_key] = self._SIM_COL

    def _final_annot_single_sim(self, df):
        return df.loc[df[self._time_key] == self._MAX_TIME][self._obs_key].values[0]

    @property
    def GROUPED_BY_FINAL_STATE(self):
        return self.adata.obs.groupby(self.key_added)

    def annotate_final_state(self, key_added="final_state"):
        self.key_added = key_added
        STATE = self._SIMULATION_GROUPED.apply(self._final_annot_single_sim)
        self.adata.obs[key_added] = self.adata.obs[self._sim_key].map(STATE)

        return self.GROUPED_BY_FINAL_STATE




__module_name__ = "__init__.py"
__doc__ = """tools __init__ module. Sub-package of the main scdiffeq API."""
__author__ = ", ".join(["Michael E. Vinyard",])
__email__ = ", ".join(["mvinyard@broadinstitute.org",])


# import functions accessed as sdq.tl.<func>: --------------------------------------------
from ._annotate_cells import annotate_cells
from ._time_free_sampling import time_free_sampling
from ._hyperparams import HyperParams
from ._reconstruct_function import reconstruct_function
from ._versions import Versions, configure_version
from ._func_from_version import func_from_version
from ._umap import UMAP
# from ._fetch import fetch
from ._drift_diffusion_state_characterization import drift, diffusion

from ._dimension_reduction import DimensionReduction


# -----------
# from ._data_formatter import DataFormatter
# from ._x_use import X_use, fetch_formatted_data
from ._knn import kNN
from ._knn_smoothing import kNNSmoothing

from ._negative_cross_entropy import NegativeCrossEntropy

from ._sum_norm_df import sum_norm_df


from ._feature_correlation import (
    FeatureCorrelation,
    drift_correlated_features,
    diffusion_correlated_features,
    potential_correlated_features,
)

from ._cell_potential import cell_potential, normalize_cell_potential
from ._norm import L2Norm

from ._final_state_per_simulation import FinalStatePerSimulation
from ._simulator import Simulator, simulate
from ._perturbation import Perturbation, perturb

from ._grouped_expression import GroupedExpression, grouped_expression
from ._annotate_gene_features import GeneCompatibility, annotate_gene_features
# from ._compared_temporal_gene_expression import TemporalGeneExpression, compared_temporal_gene_expression
# from ._smoothed_expression import SmoothedExpression, smoothed_expression
from ._temporal_expression import temporal_expression





from . import mix_ins, base


from neural_diffeqs import PotentialODE
import torch_nets
import torch


from typing import Optional, Union, List
from ... import __version__


class LightningODE_VAE_FixedPotential(
    mix_ins.BaseForwardMixIn,
    base.BaseLightningDiffEq,
    mix_ins.PreTrainMixIn,
    mix_ins.PotentialMixIn,
):
    def __init__(
        self,
        data_dim,
        latent_dim: int = 50,
        name: Optional[str] = None,
        mu_hidden: Union[List[int], int] = [2000, 2000],
        mu_activation: Union[str, List[str]] = 'LeakyReLU',
        mu_dropout: Union[float, List[float]] = 0.2,
        mu_bias: bool = True,
        mu_output_bias: bool = True,
        mu_n_augment: int = 0,
        train_lr=1e-5,
        pretrain_lr=1e-3,
        pretrain_epochs=100,
        pretrain_optimizer=torch.optim.Adam,
        train_optimizer=torch.optim.RMSprop,
        pretrain_scheduler=None,
        train_scheduler=torch.optim.lr_scheduler.StepLR,
        pretrain_step_size=None,
        train_step_size=10,
        dt=0.1,
        adjoint=False,
        backend = "auto",
        version = __version__,
        *args,
        **kwargs,
    ):
        super().__init__()
        
        name = self._configure_name(name)

        self.save_hyperparameters()

        # -- torch modules: ----------------------------------------------------
        self._configure_torch_modules(func=PotentialODE, kwargs=locals())
        self._configure_lightning_model(kwargs = locals())

    def forward(self, X0, t, **kwargs):
        """Forward step: (0) integrate in latent space"""
        Z0 = self.Encoder(X0)
        Z_hat = self.integrate(Z0=Z0, t=t, dt=self.hparams["dt"], logqp=False, **kwargs)
        return self.Decoder(Z_hat)

    def step(self, batch, batch_idx, stage=None):

        batch = self.process_batch(batch, batch_idx)
        X_hat = self.forward(batch.X0, batch.t)
        sinkhorn_loss = self.compute_sinkhorn_divergence(
            batch.X, X_hat, batch.W, batch.W_hat
        )
        return self.log_sinkhorn_divergence(sinkhorn_loss).sum()

    def pretrain_step(self, batch, batch_idx, stage=None):

        batch = self.process_batch(batch, batch_idx)
        X0_hat = self.Decoder(self.Encoder(batch.X0))
        recon_loss = self.reconstruction_loss(X0_hat, batch.X0).sum()
        self.log("pretrain_rl_mse", recon_loss.item())
        return recon_loss

    def __repr__(self):
        return "LightningODE-VAE-FixedPotential"




import yaml
import ABCParse

class HyperParams(ABCParse.ABCParse):
    def __init__(self, yaml_path):

        self.__configure__(locals())

    def _read(self):
        if not hasattr(self, "_file"):
            self._yaml_file = yaml.load(open(self._yaml_path), Loader = yaml.Loader)

    def __configure__(self, kwargs, private=["yaml_path"]):

        self.__parse__(kwargs, private=private)
        self._read()
        for key, val in self._yaml_file.items():
            setattr(self, key, val)
            
    @property
    def attrs(self):
        self._attrs = {attr: getattr(self, attr) for attr in self.__dir__() if not attr[0] in ["_", "a"]}  
        return self._attrs
    
    def __repr__(self):
        
        """Return a readable representation of the discovered hyperparameters"""
        
        
        string = "HyperParameters\n"
        for attr, val in self.attrs.items():
            string += "\n  {:<34}: {}".format(attr, val)
            
        return string


# -- experimental
# -- considering deprecating the more complicated logging
import pathlib
import pandas as pd

class LoggerBridge:
    def __init__(self, DiffEq, idx: int = 0):
        
        """ """
        
        if not len(DiffEq.loggers) < idx + 1:
            self.logger = DiffEq.loggers[idx]

    @property
    def log_dir(self):
        return pathlib.Path(self.logger.log_dir)
    
    @property
    def log_files(self):
        return list(self.log_dir.glob("*"))
        
    @property
    def metrics_csv_path(self):
        matched = [f for f in self.log_files if "metrics" in f.name]
        if matched:
            return  matched[0]
    
    @property
    def log_df(self):
        if not self.metrics_csv_path is None:
            return pd.read_csv(self.metrics_csv_path)
        
    @property
    def version(self):
        return pathlib.Path(self.log_dir).name
    




# from ... import tools

# from autodevice import AutoDevice
# NoneType = type(None)

# def fetch_format(adata, use_key: str, idx=None, N=1, device=AutoDevice()):

#     if isinstance(idx, NoneType):
#         idx = range(len(adata))

#     return tools.fetch(adata[idx], use_key=use_key, device=device)[:, None].expand(
#         -1, N, -1
#     )





# -- import packages: ---------------------------------------------------------
import sklearn
import anndata
import ABCParse
import pandas as pd


# -- import local dependencies: -----------------------------------------------
from ..core import utils


# -- set typing: --------------------------------------------------------------
from typing import Optional


# -- controller class: --------------------------------------------------------
class GeneCompatibility(ABCParse.ABCParse):
    """
    A simulation is produced in a latent space. To study gene-level
    features, we can modify the simulated adata to be annotated with
    gene-level features (which otherwise only carries latent space
    features).
    """

    def __init__(
        self,
        gene_id_key="gene_ids",
        PCA: Optional[sklearn.decomposition.PCA] = None,
        key_added: str = "X_gene",
    ):

        self.__parse__(locals(), public=[None])
        self._INFO = utils.InfoMessage()
        
        
    @property
    def var_names(self):
        return self.adata.var[self._gene_id_key]

    def _format_var_names(self):
        self.adata_sim.uns[self._gene_id_key] = self.var_names
        self._INFO(f"Gene names added to: `adata_sim.uns['{self._gene_id_key}']`")

    def _format_inverted_expression(self):

        assert not self._PCA is None, "Must supply PCA model!"
        
        X_gene = self._PCA.inverse_transform(self.adata_sim.X)
        X_gene = pd.DataFrame(
            X_gene,
            index = self.adata_sim.obs.index,
            columns = self.var_names,
        )
        self.adata_sim.obsm[self._key_added] = X_gene
        
        msg = f"Inverted expression added to: `adata_sim.obsm['{self._key_added}']`"
        self._INFO(msg)

    def __call__(
        self, adata: anndata.AnnData, adata_sim: anndata.AnnData, *args, **kwargs
    ):

        self.__update__(locals(), public = ["adata", "adata_sim"])

        self._format_var_names()
        if not self._PCA is None:
            self._format_inverted_expression()

# -- API-facing function: ------------------------------------------------------
def annotate_gene_features(
    adata_sim: anndata.AnnData,
    adata: anndata.AnnData,
    PCA: Optional[sklearn.decomposition.PCA] = None,
    gene_id_key="gene_ids",
    key_added: str = "X_gene",
):
    """
    Annotate simulation with gene-level features

    Parameters
    ----------
    adata_sim: anndata.AnnData
        Simulated AnnData object

    adata: anndata.AnnData
        Reference AnnData object

    PCA: Optional[sklearn.decomposition.PCA]
        default: None

    gene_id_key: str
        default: "gene_ids"

    key_added: str
        default: "X_gene"

    Returns
    -------
    None, modifies adata_sim in-place.
    """
    gene_compatibility = GeneCompatibility(
        gene_id_key=gene_id_key, PCA=PCA, key_added=key_added
    )
    return gene_compatibility(adata_sim=adata_sim, adata=adata)




# -- import packages: ----------------------------------------------------------
import pandas as pd
import numpy as np
import sklearn
import anndata
import umap
import ABCParse

# -- import local dependencies: ------------------------------------------------
from ..core import utils
from .. import _backend_utilities as backend

# -- set typing: ---------------------------------------------------------------
from typing import Optional, Union, List

NoneType = type(None)


# -- Operator class: -----------------------------------------------------------
class Perturbation(ABCParse.ABCParse):
    """
    Alter the relative expression of one or more genes in one
    or more cells.
    """
    def __init__(
        self,
        adata: anndata.AnnData,
        gene_id_key: str = "gene_ids",
        use_key: str = "X_scaled",
        PCA: Optional[sklearn.decomposition.PCA] = None,
        UMAP: Optional[umap.UMAP] = None,
        copy: bool = False,
    ) -> NoneType:
        
        """ """
        
        self.__parse__(locals(), ignore = ['adata'], public = [None])
        
        if self._copy:
            self.adata = adata.copy()
        else:
            self.adata = adata
        
        self.obs_lookup = backend.ObsIndexLookUp()
        self.var_lookup = backend.VarIndexLookUp(var_name_key = gene_id_key)

    @property
    def X(self) -> np.ndarray:
        if not hasattr(self, "_X"):
            self._X = self.adata.layers[self._use_key].copy()
        return self._X

    @property
    def X_pca_perturbed(self):
        if not hasattr(self, "_X_pca_perturbed"):
            self._X_pca_perturbed = self._PCA.transform(self.X_perturb[self.cell_idx])
        return self._X_pca_perturbed

    @property
    def X_umap_perturbed(self):
        if not hasattr(self, "_X_umap_perturbed"):
            self._X_umap_perturbed = self._UMAP.transform(self.X_pca_perturbed)
        return self._X_umap_perturbed

    @property
    def obs_perturbed(self):
        perturb_indicator = np.zeros(self.adata.shape[0], dtype=bool)
        perturb_indicator[self.cell_idx] = True
        return perturb_indicator

    @property
    def var_perturbed(self):
        perturb_indicator = np.zeros(self.adata.shape[1], dtype=bool)
        perturb_indicator[self.gene_idx] = True
        return perturb_indicator
        
    def _update_adata(self):
        
        self.adata.uns["X_perturb"] = self.X_perturb[self.cell_idx]
        
        if not isinstance(self._PCA, NoneType):
            self.adata.uns["X_pca_perturb"] = self.X_pca_perturbed
        if not isinstance(self._UMAP, NoneType):
            self.adata.uns["X_umap_perturb"] = self.X_umap_perturbed
        
        self.adata.obs["perturbed"] = self.obs_perturbed
        self.adata.var["perturbed"] = self.var_perturbed
        
        self.adata.uns['perturbed'] = {
            "genes": self._gene_idx,
            "cells": self._cell_idx,
        }
        
        
    def forward(self, cell_idx: Union[int, List], gene_idx: int, val: float):
        """ """
        self.X_perturb[cell_idx, gene_idx] = val
        
    @property
    def cell_idx(self):
        return self.obs_lookup(self.adata, self._cell_idx)
    
    @property
    def gene_idx(self):
        return self.var_lookup(self.adata, self._gene_idx)
    
    def _prepare_return(self):
        
        return_ = []
        if self.return_array:
            return_.append(self.X_perturb)
        if self._copy:
            return_.append(self.adata)
        
        if len(return_) > 0:
            return return_
        
        
    def __call__(
        self,
        cell_idx: Union[int, List],
        gene_idx: List,
        val: float,
        return_array: bool = False,
    ):
        """ """
        
        self.__update__(locals(), private = ['cell_idx', 'gene_idx'])
        
        self.X_perturb = self.X
        
        for gene_ix in self.gene_idx:
            self.forward(self.cell_idx, gene_ix, val)
        
        self._update_adata()
            
        return self._prepare_return()


# -- API-facing function: ------------------------------------------------------
def perturb(
    adata: anndata.AnnData,
    cell_idx: Union[pd.Index, np.array, list],
    gene_idx: Union[pd.Index, np.array, list],
    val: float,
    use_key: str = "X_scaled",
    PCA: Optional[sklearn.decomposition.PCA] = None,
    UMAP: Optional[umap.UMAP] = None,
    return_array: bool = False,
    copy: bool = False,
) -> anndata.AnnData:
    
    """
    Impart a perturbation.
    
    Parameters:
    -----------
    adata
        Annotated Data object.
        type: anndata.AnnData
        
    cell_idx
        Cells to be perturbed. Should correspond to the integer indices of the layer
        chosen by adata.layers[use_key].
        type: Union[pd.Index, np.array, list]
        
    gene_idx
        Genes to be perturbed. Should correspond to the integer indices of the layer
        chosen by adata.layers[use_key].
        type: Union[pd.Index, np.array, list]
    
    val
        Modified expression value
        type: float

    use_key
        Layer from which reference gene expression should be taken.
        type: str
        default: "X_scaled"
        
    PCA
        PCA model
        type: Optional[sklearn.decomposition.PCA]
        default: None
        
    UMAP
        UMAP model
        type: Optional[umap.UMAP]
        default: None
        
    copy
        If `True`, returns a copy of the input `adata`.
        type: bool
        default: False
        
    return_array
        type: bool
        default: False
        

    Returns:
    --------
    adata_perturbed
    """
    
    perturbation = Perturbation(
        adata = adata,
        PCA = PCA,
        UMAP = UMAP,
        use_key = use_key,
        copy = copy,
    )
    return perturbation(
        cell_idx = cell_idx,
        gene_idx = gene_idx,
        val = val,
        return_array = return_array,
    )
    



from neural_diffeqs import LatentPotentialODE
import torch

from . import base, mix_ins

from .. import utils

from typing import Optional, Union, List
from ... import __version__


class LightningODE_PriorPotential(
    base.BaseLightningDiffEq,
    
):
    def __init__(
        self,
        # -- general params: ---------------------------------------------------
        latent_dim: int = 50,
        name: Optional[str] = None,
        train_lr=1e-5,
        train_optimizer=torch.optim.RMSprop,
        train_scheduler=torch.optim.lr_scheduler.StepLR,
        train_step_size=10,
        adjoint=False,
        
        # -- ODE params: -------------------------------------------------------
        coef_diff: float = 0,
        dt: float = 0.1,
        mu_hidden: Union[List[int], int] = [400, 400],
        mu_activation: Union[str, List[str]] = 'LeakyReLU',
        mu_dropout: Union[float, List[float]] = 0.2,
        mu_bias: bool = True,
        mu_output_bias: bool = True,
        mu_n_augment: int = 0,
        sde_type='ito',
        noise_type='general',
        backend = "auto",
        brownian_dim=1,
        
        version = __version__,
        
        *args,
        **kwargs,
    ):
        super().__init__()
        
        name = self._configure_name(name)

        self.save_hyperparameters()        
        self.func = LatentPotentialODE(
            state_size=latent_dim,
            **utils.extract_func_kwargs(func=LatentPotentialODE, kwargs=locals()),
        )
        self._configure_lightning_model(kwargs = locals())
        
    def forward(self, X0, t, **kwargs):
        """Forward step: (0) integrate in latent space"""
        Z_hat, KL_div = self.integrate(
            Z0=X0, t=t, dt=self.hparams["dt"], logqp=True, **kwargs
        )
        return Z_hat, KL_div

    def log_computed_loss(self, sinkhorn_loss, t, kl_div_loss, stage):
        
        sinkhorn_loss = self.log_sinkhorn_divergence(sinkhorn_loss).sum()
        self.log(f"kl_div_{stage}", kl_div_loss.sum())

        return sinkhorn_loss + kl_div_loss.sum()

    def step(self, batch, batch_idx, stage=None):

        batch = self.process_batch(batch, batch_idx)
        X_hat, kl_div_loss = self.forward(batch.X0, batch.t)
        sinkhorn_loss = self.compute_sinkhorn_divergence(
            batch.X, X_hat, batch.W, batch.W_hat
        )
        return self.log_computed_loss(
            sinkhorn_loss, t=batch.t, kl_div_loss=kl_div_loss, stage=stage
        )

        
    def __repr__(self):
        return "LightningODE-PriorPotential"



# -- import packages: ----------------------------------------------------------
import licorice_font


# -- set typing: ---------------------------------------------------------------
NoneType = type(None)


class InfoMessage:
    def __init__(
        self, silent: bool = False, INFO: str = "INFO", color: str = "BLUE"
    ) -> NoneType:

        """

        Parameters:
        -----------
        silent
            type: bool
            default: False

        INFO
            type: str
            default: "INFO"

        color
            type: str
            default: "BLUE"

        Returns:
        --------
        None
        """

        self.INFO = licorice_font.font_format(INFO, [color])
        self._SILENT = silent

    def _format_msg_title(self, INFO) -> str:
        return f" - [{INFO}] | "

    def __call__(self, msg: str) -> NoneType:

        """
        Prints message according to above-specified formatting.

        Parameters:
        -----------
        msg
            type: str
        """

        if not self._SILENT:
            self.BASE = self._format_msg_title(self.INFO)
            print(self.BASE + msg)




import os

class _PackageVersion:
    
    def __init__(self):
        ...
        
    @property
    def PACKAGE_PATH(self):
        return _os.path.abspath(_os.path.dirname(_os.path.dirname(__file__)))
    
    @property
    def SETUP_FPATH(self):
        return _os.path.join(self.PACKAGE_PATH, "setup.py")
    
    def _read_setup_dot_py(self):
        f = open(self.SETUP_FPATH)
        file = f.readlines()
        f.close()
        return file
    
    @property
    def VERSION(self):
        file = self._read_setup_dot_py()
        return [line for line in file if "version" in line][0].split("version=")[1].split('"')[1]
    
    def __call__(self):
        return self.VERSION



from typing import Optional, Union, List, Dict
import numpy as np
import sklearn
import anndata
import ABCParse

from ..core import utils
from ._annotate_gene_features import annotate_gene_features
from ._grouped_expression import grouped_expression


class TemporalGeneExpression(ABCParse.ABCParse):
    def __init__(
        self,
        adata: anndata.AnnData,
        adata_sim: anndata.AnnData,
        gene_id_key="gene_ids",
        sim_time_key="t",
        ref_time_key="Time point",
        sim_use_key="X_gene",
        ref_use_key="X_scaled",
        PCA: Optional[sklearn.decomposition.PCA] = None,
    ):
        self.__parse__(locals(), ignore=["adata", "adata_sim", "PCA"], public=[None])

        self.adata_sim = adata_sim.copy()
        self.adata = adata.copy()
        self.adata.var_names = adata.var[self._gene_id_key]

        annotate_gene_features(self.adata_sim, self.adata, PCA=PCA)

    @property
    def REF_GEX(self):
        return grouped_expression(
            self.adata,
            gene_id=self._gene_id,
            groupby=self._ref_time_key,
            use_key=self._ref_use_key,
        )

    @property
    def SIM_GEX(self):
        return grouped_expression(
            self.adata_sim,
            gene_id=self._gene_id,
            groupby=self._sim_time_key,
            use_key=self._sim_use_key,
        )

    def __call__(self, gene_id: Union[List[str], str], *args, **kwargs) -> Dict:

        self.__update__(locals(), private=["gene_id"])

        return {
            "ref": self.REF_GEX,
            "sim": self.SIM_GEX,
        }


def compared_temporal_gene_expression(
    adata: anndata.AnnData,
    adata_sim: anndata.AnnData,
    gene_id: Union[List[str], str],
    gene_id_key: str = "gene_ids",
    sim_time_key: str = "t",
    ref_time_key: str = "Time point",
    sim_use_key: str = "X_gene",
    ref_use_key: str = "X_scaled",
    PCA: Optional[sklearn.decomposition.PCA] = None,
    *args,
    **kwargs
) -> Dict:
    
    """
    Compare gene expression across pseudo-time-steps and real time.
    
    Parameters
    ----------
    adata: anndata.AnnData,
    
    adata_sim: anndata.AnnData,
    
    gene_id: Union[List[str], str],
    
    gene_id_key: str = "gene_ids",
    
    sim_time_key: str = "t",
    
    ref_time_key: str = "Time point",
    
    sim_use_key: str = "X_gene",
    
    ref_use_key: str = "X_scaled",
    
    PCA: Optional[sklearn.decomposition.PCA] = None,
    
    Returns
    -------
    compared_gex: Dict
    
    Notes
    -----
    
    """
    time_gex = TemporalGeneExpression(
        adata=adata,
        adata_sim=adata_sim,
        gene_id_key=gene_id_key,
        sim_time_key=sim_time_key,
        ref_time_key=ref_time_key,
        sim_use_key=sim_use_key,
        ref_use_key=ref_use_key,
        PCA=PCA,
    )
    return time_gex(gene_id=gene_id)



import ABCParse
import os

from .. import lightning_models, utils


NoneType = type(None)
# figure out how to return model from file wihtout doing anyhhing else - fast, easy
#     def load_from_ckpt(self):
        
#         return 
#         self.DiffEq = 
#         def load_model(self):
#         diffeq = sdq.core.lightning_models.LightningSDE_FixedPotential.

# -- import packages: ----------------------------------------------------------
class LightningModelConfiguration(ABCParse.ABCParse):
    _potential_types = {
        "fixed": "FixedPotential",
        "prior": "PriorPotential",
    }

    def __init__(
        self,
        data_dim,
        latent_dim: int = 20,
        DiffEq_type: str = "SDE",
        potential_type="prior",
        fate_bias_csv_path = None,
    ):

        self.__parse__(locals(), public=[None])
        


    @property
    def available_lightning_models(self):
        return [
            attr
            for attr in lightning_models.__dir__()
            if attr.startswith("Lightning")
        ]

    @property
    def DiffEq_type(self):
        return "".join(["Lightning", self._DiffEq_type])

    @property
    def use_vae(self):
        return self._data_dim > self._latent_dim

    @property
    def potential_type(self):
        if self._potential_type:
            return self._potential_types[self._potential_type]
        
    @property
    def fate_bias_aware(self):
        if isinstance(self._fate_bias_csv_path, NoneType):
            return False
        else:
            if os.path.exists(self._fate_bias_csv_path):
                return True
            else:
                raise ValueError("Path to fate_bias.csv was passed though not found.")

    @property
    def _USE_CKPT(self):
        return isinstance(self._ckpt_path, str)
    
    def __call__(self, kwargs, ckpt_path = None):
        
        self._ckpt_path = ckpt_path

        _model = [self.DiffEq_type]

        if self.use_vae:
            _model.append("VAE")

        if self.potential_type:
            _model.append(self.potential_type)
            
        if self.fate_bias_aware:
            _model.append("FateBiasAware")

        _model = "_".join(_model)

        if _model in self.available_lightning_models:
            lit_model = getattr(lightning_models, _model)
            
            if self._USE_CKPT:
                return lit_model.load_from_checkpoint(self._ckpt_path)
            
            model_kwargs = utils.function_kwargs(func=lit_model.__init__, kwargs=kwargs)
            
            return lit_model(data_dim = self._data_dim, **model_kwargs)
        
        raise ValueError(f"Configuration tried: {_model} - this does not exist as an available model.")




from ._info_message import InfoMessage
from ._anndata_inspector import AnnDataInspector

import annoyance
import pandas as pd
import torch
import numpy as np

NoneType = type(None)

class FastGraph:
    def __init__(self, adata, use_key, annot_key="Cell type annotation"):

        self._INFO = InfoMessage()

        self.adata = adata
        self._inspector = AnnDataInspector(adata)
        self.annot_key = annot_key
        self.use_key = use_key
        if use_key in self._inspector.layers:
            adata.obsm[use_key] = adata.layers[use_key]

        self.Graph = annoyance.kNN(adata, use_key=self.use_key)
        self.Graph.build()

    def _fast_count(self, X_nn):
        return (
            self.adata[X_nn.flatten().astype(str)]
            .obs[self.annot_key]
            .values.reshape(-1, self.Graph._n_neighbors)
        )

    def _query(self, X_fin):
        return self._fast_count(self.Graph.query(X_fin))

    def _fate_df(self, x_lab):
        return pd.DataFrame([x_lab[0].value_counts() for i in range(len(x_lab))])
    
    def _DETACH(self, X_hat: torch.Tensor)->np.ndarray:
        return X_hat.detach().cpu().numpy()
    
    def _DETACH_FINAL(self, X_hat: torch.Tensor)->np.ndarray:
        return X_hat[-1].detach().cpu().numpy()
    
    def _TRANSFORM(self, dimension_reduction_model, X_fin):
        return dimension_reduction_model.transform(X_fin)
        

    def __call__(self, X_hat, dimension_reduction_model = None, final_timepoint_only=True):
        
        if X_hat.device != "cpu" and (final_timepoint_only):
            X_hat_ = self._DETACH_FINAL(X_hat)
        elif X_hat.device != "cpu":
            X_hat_ = self._DETACH(X_hat)
        elif final_timepoint_only:
            X_hat_ = X_hat[-1].numpy()
        else:
            X_hat_ = X_hat.numpy()
            
        if not isinstance(dimension_reduction_model, NoneType):
            X_hat_ = self._TRANSFORM(dimension_reduction_model, X_hat_)
        return self._fate_df(self._query(X_fin))




# -- import packages: ----------------------------------------------------------
from neural_diffeqs import NeuralODE
import torch_nets
import torch


# -- import local dependencies: ------------------------------------------------
from . import mix_ins
from . import base


from typing import Optional, Union, List
from ... import __version__



# -- lightning model: ----------------------------------------------------------
class LightningODE_VAE(
    mix_ins.VAEMixIn,
    mix_ins.PreTrainMixIn,
    base.BaseLightningDiffEq,
):
    def __init__(
        self,
        data_dim: int,
        latent_dim: int = 50,
        name: Optional[str] = None,
        train_lr=1e-5,
        pretrain_lr=1e-3,
        pretrain_epochs=100,
        pretrain_optimizer=torch.optim.Adam,
        train_optimizer=torch.optim.RMSprop,
        pretrain_scheduler=torch.optim.lr_scheduler.StepLR,
        train_scheduler=torch.optim.lr_scheduler.StepLR,
        pretrain_step_size=200,
        train_step_size=10,
        dt=0.1,
        adjoint=False,
        backend = "auto",

        # -- ode params: -------------------------------------------------------
        mu_hidden: Union[List[int], int] = [400, 400, 400],
        mu_activation: Union[str, List[str]] = 'LeakyReLU',
        mu_dropout: Union[float, List[float]] = 0.1,
        mu_bias: bool = True,
        mu_output_bias: bool = True,
        mu_n_augment: int = 0,
        sde_type='ito',
        noise_type='general',
        
        # -- encoder parameters: -----------------------------------------------
        encoder_n_hidden: int = 4,
        encoder_power: float = 2,
        encoder_activation: Union[str, List[str]] = 'LeakyReLU',
        encoder_dropout: Union[float, List[float]] = 0.2,
        encoder_bias: bool = True,
        encoder_output_bias: bool = True,
        
        # -- decoder parameters: -----------------------------------------------
        decoder_n_hidden: int = 4,
        decoder_power: float = 2,
        decoder_activation: Union[str, List[str]] = 'LeakyReLU',
        decoder_dropout: Union[float, List[float]] = 0.2,
        decoder_bias: bool = True,
        decoder_output_bias: bool = True,
        version = __version__,
        *args,
        **kwargs,
    ):
        super().__init__()
        
        name = self._configure_name(name)

        self.save_hyperparameters()
        
        # -- torch modules: ----------------------------------------------------
        self._configure_torch_modules(func = NeuralODE, kwargs=locals())
        self._configure_lightning_model(kwargs = locals())

    def __repr__(self):
        return "LightningODE-VAE"



# -- import packages: ----------------------------------------------------------
from geomloss import SamplesLoss
import autodevice
import ABCParse


# -- API-facing class: ---------------------------------------------------------
class SinkhornDivergence(SamplesLoss, ABCParse.ABCParse):
    def __init__(
        self,
        device=autodevice.AutoDevice(),
        loss="sinkhorn",
        backend="auto",
        p=2,
        blur=0.1,
        scaling=0.7,
        debias=True,
        sample_axis=1,
        reach=None,
        diameter=None,
        truncate=5,
        cost=None,
        kernel=None,
        cluster_scale=None,
        potentials=False,
        verbose=False,
        **kwargs
    ):
        super(SinkhornDivergence, self).__init__()

        self.__parse__(locals())




class BaseForwardMixIn(object):
    def __init__(self):
        super().__init__()


    def forward(self, Z0, t, **kwargs):
        """Forward step: (0) integrate in latent space"""
        return self.integrate(Z0=Z0, t=t, dt=self.hparams["dt"], logqp=False, **kwargs)
    
    def step(self, batch, batch_idx, stage=None):

        batch = self.process_batch(batch, batch_idx)
#         print(batch_idx, batch.X0.mean())
        X_hat = self.forward(batch.X0, batch.t)
        
#         print("X:", batch.X.shape)
#         print("X_hat:", X_hat.shape)
#         print("W (shape, first five):", batch.W.shape, batch.W[2][:5])
#         print("W_hat (shape, first five):", batch.W_hat.shape, batch.W_hat[2][:5])
        
        self.sinkhorn_loss = self.compute_sinkhorn_divergence(
            X = batch.X, X_hat = X_hat, W = batch.W, W_hat = batch.W_hat
        )
#         print("batch_loss", self.sinkhorn_loss.sum())
        return self.sinkhorn_loss.sum()




# -- import packages: ---------------------------------------------------------
import anndata
import ABCParse
import matplotlib.pyplot as plt
import cellplots as cp
import matplotlib.cm as cm


# -- set typing: --------------------------------------------------------------
from typing import Optional, List, Dict, Union


# -- Operational class: -------------------------------------------------------
class TemporalExpressionPlot(ABCParse.ABCParse):
    def __init__(
        self,
        height: float = 0.4,
        width: float = 0.4,
        delete: List[List[str]] = [["top", "right"]],
        plot_kwargs: Dict = {},
        *args,
        **kwargs,
    ):
        self.__parse__(locals(), public=[None])

    @property
    def _MEAN(self):
        return self._adata_sim.uns["gex_mean"][self._gene]

    @property
    def _STD(self):
        return self._adata_sim.uns["gex_std"][self._gene]

    @property
    def _HI(self):
        return self._MEAN + self._STD

    @property
    def _LO(self):
        _lo = self._MEAN - self._STD
        _lo[_lo < 0] = 0
        return _lo

    @property
    def _columns(self):
        return self._MEAN.columns.tolist()

    @property
    def _X(self):
        return self._MEAN.index.to_numpy().round(2)

    @property
    def _global_max(self):
        return self._HI.max().max()

    @property
    def _PLOT_GROUPS(self):
        if not hasattr(self, "_plot_groups"):
            return self._columns
        return self._plot_groups

    @property
    def ax(self):
        if not hasattr(self, "_ax"):
            fig, axes = cp.plot(
                1,
                1,
                height=self._height,
                width=self._width,
                delete=self._delete,
                **self._plot_kwargs,
            )
            self._ax = axes[0]
        return self._ax

    @property
    def _CMAP(self):
        if not hasattr(self, "_cmap"):
            self._cmap = cm.tab20.colors
        return self._cmap

    def _fetch_color(self, en, group):

        if group in self._CMAP:
            return self._CMAP[group]
        return self._CMAP[en]

    def _plot_std(self, group, color):

        y1 = self._LO[group]
        y2 = self._HI[group]
        self.ax.fill_between(
            x=self._X,
            y1=y1,
            y2=y2,
            alpha=0.05,
            ec="None",
            linewidth=0,
            color=color,
        )

    @property
    def _sim_idx(self):
        return self._adata_sim.uns["idx"]

    def forward(self, en, group):

        color = self._fetch_color(en, group)

        if self._show_std:
            self._plot_std(group, color)

        x, y = self._X, self._MEAN[group]
        self.ax.plot(x, y, lw=2, color=color, zorder=en + 2, label=group)
        self.ax.set_xlim(x.min(), x.max())
#         self.ax.set_ylim(0, round(self._global_max))

    def __call__(
        self,
        adata_sim: anndata.AnnData,
        gene: str,
        plot_groups: Optional[List[str]] = None,
        show_std: bool = False,
        ax: Optional[plt.Axes] = None,
        cmap: Optional[Dict] = None,
        save: bool = False,
        savename: str = None,
        save_format: str = "svg",
        dpi: int = 250,
        x_label: str = "t (d)",
        y_label: str = None,
        title: str = None,
        title_fontsize: int = 8,
        grid: bool = True,
        show_legend: bool = True,
        *args,
        **kwargs,
    ):
        self.__update__(locals(), public=[None])

        for en, group in enumerate(self._PLOT_GROUPS):
            self.forward(en, group)

        self.ax.set_xlabel(self._x_label, fontsize=8)
        
        if y_label is None:
            y_label = self._gene
        if title is None:
            title = f"sim idx: {self._sim_idx}"
        self.ax.set_ylabel(y_label, fontsize=8)    
        self.ax.set_title(title, fontsize=title_fontsize)
            
        if grid:
            self.ax.grid(zorder=0, alpha=0.1)
        
        if show_legend:
            self.ax.legend(edgecolor="None", facecolor="None", loc=(1.05, 0.25), fontsize=6)
        

        if save:
            if savename is None:
                savename = f"scDiffEq.temporal_expression.sim_idx_{self._sim_idx}.{gene}.{save_format}"
            plt.savefig(savename, dpi=dpi)


# -- API-facing function: ----------------------------------------------------- 
def temporal_expression(
    adata_sim: anndata.AnnData,
    gene: str,
    plot_groups: Optional[List[str]] = None,
    show_std: bool = False,
    ax: Optional[plt.Axes] = None,
    cmap: Optional[Dict] = None,
    height: float = 0.4,
    width: float = 0.4,
    delete: List[List[str]] = [["top", "right"]],
    save: bool = False,
    savename: str = None,
    save_format: str = "svg",
    dpi: int = 250,
    x_label: str = "t (d)",
    y_label: str = None,
    title: str = None,
    grid: bool = True,
    show_legend: bool = True,
    plot_kwargs: Optional[Dict] = {},
    title_fontsize: int = 8,
):
    """
    Plot smoothed expression over time.
    
    Parameters
    ----------
    adata_sim: anndata.AnnData
    
    gene: str
    
    plot_groups: Optional[List[str]] = None
    
    show_std: bool = False
    
    ax: Optional[plt.Axes] = None
    
    cmap: Optional[Dict] = None
    
    height: float, default = 0.4
    
    width: float, default = 0.4
    
    delete: List[List[str]], default = [["top", "right"]]
    
    save: bool, default = False
        
    savename: str, default = None
    
    save_format: str, default = "svg"
        ["svg", "png"]
    
    dpi: int, default = 250
    
    plot_kwargs: Optional[Dict], default = {}
    
    Returns
    -------
    None
    """

    time_expr = TemporalExpressionPlot(
        height=height, width=width, delete=delete, plot_kwargs=plot_kwargs
    )
    time_expr(
        adata_sim=adata_sim,
        gene=gene,
        plot_groups=plot_groups,
        show_std=show_std,
        ax=ax,
        cmap=cmap,
        save=save,
        savename=savename,
        save_format=save_format,
        dpi=dpi,
        x_label = x_label,
        y_label = y_label,
        title = title,
        grid = grid,
        show_legend = show_legend,
        title_fontsize = title_fontsize,
    )




# -- import: -----------------------------------------------------------------------------
# from ._loss_accounting import LossAccounting
from ._intermittent_saves import IntermittentSaves
from ._gradient_potential_callback import GradientPotentialTest
from ._testing import Testing

from ._visualize_tracked_loss import ModelTracker, VisualizeTrackedLoss
from ._visualize_predictions import VisualizePredictions
from ._model_logging import ModelLogging



# -- import packages: ----------------------------------------------------------
import torch_nets
import lightning
import torch


from ... import utils

# -- set typing: ---------------------------------------------------------------
from typing import Union, List
NoneType = type(None)


# -- MixIn class: --------------------------------------------------------------        
class VAEMixIn(object):
    def __init__(self, *args, **kwargs):
        super().__init__()

        # -- loss functions: ---------------------------------------------------
        self.reconstruction_loss = torch.nn.MSELoss(reduction="sum", **kwargs)
        
    def _configure_encoder(self):
        self.Encoder = torch_nets.Encoder(
            data_dim = self.hparams['data_dim'],
            latent_dim = self.hparams['latent_dim'],
            n_hidden=self.hparams["encoder_n_hidden"],
            power=self.hparams["encoder_power"],
            activation=self.hparams["encoder_activation"],
            dropout=self.hparams["encoder_dropout"],
            bias=self.hparams["encoder_bias"],
            output_bias=self.hparams["encoder_output_bias"],
        )

    def _configure_decoder(self):
        self.Decoder = torch_nets.Decoder(
            data_dim = self.hparams['data_dim'],
            latent_dim = self.hparams['latent_dim'],
            n_hidden=self.hparams["decoder_n_hidden"],
            power=self.hparams["decoder_power"],
            activation=self.hparams["decoder_activation"],
            dropout=self.hparams["decoder_dropout"],
            bias=self.hparams["decoder_bias"],
            output_bias=self.hparams["decoder_output_bias"],
        )
        
    def _configure_torch_modules(self, func, kwargs):
        
        kwargs['state_size'] = self.hparams['latent_dim']
        
        self._configure_encoder()
        self.DiffEq = func(**utils.function_kwargs(func, kwargs))
        self._configure_decoder()
        
    def _configure_optimizers_schedulers(self):
        
        """Assumes use of a pre-train step with an independent optimizer, scheduler (i.e., two total, each)"""
        
        pretrain_optimizer = self.hparams['pretrain_optimizer']
        train_optimizer = self.hparams['train_optimizer']
        pretrain_scheduler = self.hparams['pretrain_scheduler']
        train_scheduler = self.hparams['train_scheduler']
        
        VAE_params = list(self.Encoder.parameters()) + list(self.Decoder.parameters())
        self._optimizers = [
            pretrain_optimizer(VAE_params, lr=self.hparams['pretrain_lr']),
            train_optimizer(self.parameters(), lr=self.hparams['train_lr']),
        ]
        
        self._schedulers = [
            pretrain_scheduler(
                optimizer=self._optimizers[0], step_size=self.hparams['pretrain_step_size']
            ),
            train_scheduler(
                    optimizer=self._optimizers[1], step_size=self.hparams['train_step_size']
                ),
        ]
        
    def encode(self, X: torch.Tensor) -> torch.Tensor:
        return self.Encoder(X)

    def decode(self, Z: torch.Tensor) -> torch.Tensor:
        return self.Decoder(Z)

    def forward(self, X0, t, logqp=False, **kwargs):
        """Forward step: (0) integrate in latent space"""
        Z_hat = self.integrate(
            Z0=self.encode(X0),
            t=t,
            dt=self.hparams["dt"],
            logqp=logqp,
            **kwargs,
        )
        return self.decode(Z_hat)

    def pretrain_step(self, batch, batch_idx, stage=None):
        """Encode only X0->Z0. Return X0_hat"""
        
        pretrain_optim = self.optimizers()[0]
        pretrain_optim.zero_grad()
        batch = self.process_batch(batch, batch_idx)
        X0_hat = self.decode(self.encode(batch.X0))
        recon_loss = self.reconstruction_loss(X0_hat, batch.X0).sum()
        self.log("pretrain_rl_mse", recon_loss.item())
        
        self.log_lr()
        
        self.manual_backward(recon_loss)
        pretrain_optim.step()
        pretrain_scheduler = self.lr_schedulers()[0]
        pretrain_scheduler.step()

    def step(self, batch, batch_idx, stage=None):
        
        if stage == "training":
            train_optim = self.optimizers()[1]
            train_optim.zero_grad()

        batch = self.process_batch(batch, batch_idx)
        X_hat = self.forward(batch.X0, batch.t)
        sinkhorn_loss = self.compute_sinkhorn_divergence(
            batch.X, X_hat, batch.W, batch.W_hat
        )
        loss = self.log_sinkhorn_divergence(sinkhorn_loss, t=batch.t, stage=stage)
                
        self.log_lr()
        
        if stage == "training":
            self.manual_backward(loss)
            train_optim.step()
            train_scheduler = self.lr_schedulers()[1]
            train_scheduler.step()

    def __repr__(self):
        return "scDiffEq MixIn: VAEMixIn"





# -- import packages: ----------------------------------------------------------
import lightning
import torchsde
from abc import abstractmethod
import ABCParse

# -- import local dependencies: ------------------------------------------------
from ._batch_processor import BatchProcessor
from ._sinkhorn_divergence import SinkhornDivergence

from ... import utils


from typing import Optional

# -- DiffEq class: -------------------------------------------------------------
class BaseLightningDiffEq(lightning.LightningModule):
    def __init__(self, *args, **kwargs):
        super().__init__()
        
        
    def _update_lit_diffeq_hparams(self, model_params):
        for key, val in self.hparams.items():
            if key in model_params.keys():
                if val != model_params[key]:
                    self.hparams.update({key: model_params[key]})
        
    # -- setup: ----------------------------------------------------------------
    def _configure_lightning_model(self, kwargs):
        
        """Assumes no pre-train - i.e., a single optimizer, scheduler"""
        
        optimizer = self.hparams['train_optimizer']
        scheduler = self.hparams['train_scheduler']

        self._optimizers = [optimizer(self.parameters(), lr=self.hparams['train_lr'])]
        self._schedulers = [
            scheduler(
                optimizer=self._optimizers[0],
                step_size=self.hparams['train_step_size']),
        ]
        sinkhorn_kwargs = ABCParse.function_kwargs(func = SinkhornDivergence, kwargs = kwargs)
        self.sinkhorn_divergence = SinkhornDivergence(**sinkhorn_kwargs)
        self.process_batch = BatchProcessor
        self.COMPLETED_EPOCHS = 0

    def _configure_torch_modules(self, func, kwargs):
        
        kwargs['state_size'] = self.hparams['latent_dim']
        self.DiffEq = func(**utils.function_kwargs(func, kwargs))
        
    @property
    def PRETRAIN(self):
        return False
    
    # -- IVP-solving: ---------------------------------------------------------
    @property
    def _INTEGRATOR(self):
        if self.hparams["adjoint"]:
            return torchsde.sdeint_adjoint
        return torchsde.sdeint

    def integrate(self, Z0, t, dt, logqp, **kwargs):
        return self._INTEGRATOR(
            sde=self.DiffEq,
            y0=Z0,
            ts=t,
            dt=dt,
            logqp=logqp,
            **kwargs,
        )

    # -- sinkhorn loss: -------------------------------------------------------
    def compute_sinkhorn_divergence(self, X, X_hat, W, W_hat):
        return self.sinkhorn_divergence(
            W_hat.contiguous(), X_hat.contiguous(), W.contiguous(), X.contiguous(), 
        ).requires_grad_()


    # -- custom steps: -------------------------------------------------------------
    @abstractmethod
    def forward(self, Z0, t, **kwargs):
        """most likely over-written in another class"""
        ...

    @abstractmethod
    def step(self, batch, batch_idx, stage=None):
        print("WARNING: The base (empty) step is being called from `_base_lightning_diffeq.py`")
        ...

    # -- LightningModule methods: ----------------------------------------------
    def training_step(self, batch, batch_idx, *args, **kwargs):
        return self.step(batch, batch_idx, stage="training")

    def validation_step(self, batch, batch_idx=None, *args, **kwargs):
        return self.step(batch, batch_idx, stage="validation")

    def test_step(self, batch, batch_idx=None, *args, **kwargs):
        return self.step(batch, batch_idx, stage="test")

    def predict_step(self, batch, batch_idx=None, *args, **kwargs):
        return self.step(batch, batch_idx, stage="predict")

    def configure_optimizers(self):
        return self._optimizers, self._schedulers
    
    def __repr__(self):
        return "LightningDiffEq"
    
    def _configure_name(self, name: Optional[str] = None):
        """ """
        if not name is None:
            return f"{self.__repr__()}:{name}"
        else:
            return self.__repr__()

    
    
# -- moved to log callback: ---
#     def log_sinkhorn_divergence(self, sinkhorn_loss, t, stage):
#         for i in range(len(t)):
#             _t = round(t[i].item(), 3)
#             msg = f"sinkhorn_{_t}_{stage}"
#             val = sinkhorn_loss[i]
#             self.log(msg, val)

#         return sinkhorn_loss.sum()
    
#     def log_lr(self):
                
#         if not isinstance(self.optimizers(), list):
#             lr = self.optimizers().optimizer.state_dict()["param_groups"][0]["lr"]
#             self.log("opt_param_group_lr", lr)
#         else:
#             for i, opt in enumerate(self.optimizers()):
#                 for j, pg in enumerate(opt.optimizer.state_dict()["param_groups"]):
#                     self.log(f"opt_{i}_param_group_{j}_lr", pg["lr"])

#     def log_total_epochs(self):
#         """Train model N times --> N"""
#         self.log("total_epochs", self.COMPLETED_EPOCHS)
        




import ABCParse
import anndata
import pathlib
import lightning
import torch
import torch_nets
import yaml


from ..core import lightning_models, utils, scDiffEq


from typing import Optional, Union


class Project(ABCParse.ABCParse):
    def __init__(self, path=pathlib.Path("./").absolute(), *args, **kwargs):

        self.__parse__(locals())

        for k, v in self._VERSION_PATHS.items():
            setattr(self, k, v)

    @property
    def _PROJECT_PATH(self):
        if isinstance(self._path, pathlib.Path):
            return self._path
        elif isinstance(self._path, str):
            return pathlib.Path(self._path)
        else:
            raise TypeError("arg: `path` must be of type: [pathlib.Path, str]")

    @property
    def _VERSION_PATHS(self):
        version_paths = sorted(list(self._PROJECT_PATH.glob("version_*")))
        return {path.name: path for path in version_paths}


class HyperParams(ABCParse.ABCParse):
    def __init__(self, yaml_path):

        self.__configure__(locals())

    def _read(self):
        if not hasattr(self, "_file"):
            self._yaml_file = yaml.load(open(self._yaml_path), Loader=yaml.Loader)

    def __configure__(self, kwargs, private=["yaml_path"]):

        self.__parse__(kwargs, private=private)
        self._read()
        for key, val in self._yaml_file.items():
            setattr(self, key, val)

    @property
    def _ATTRS(self):
        self._attrs = {
            attr: getattr(self, attr)
            for attr in self.__dir__()
            if not attr[0] in ["_", "a"]
        }
        return self._attrs

    def __repr__(self):

        """Return a readable representation of the discovered hyperparameters"""

        string = "HyperParameters\n"
        for attr, val in self._ATTRS.items():
            string += "\n  {:<34}: {}".format(attr, val)

        return string

    def __call__(self):
        return self._ATTRS


class Checkpoint(ABCParse.ABCParse):
    def __init__(self, path: pathlib.Path, *args, **kwargs):
        self.__parse__(locals(), public=["path"])

    @property
    def _fname(self):
        return self.path.name.split(".")[0]

    @property
    def epoch(self):
        if self._fname != "last":
            return int(self._fname.split("=")[1].split("-")[0])
        return self._fname

    @property
    def state_dict(self):
        if not hasattr(self, "_ckpt"):
            self._state_dict = torch.load(self.path)  # ["state_dict"]
        return self._state_dict

    def __repr__(self):
        return f"ckpt epoch: {self.epoch}"


class Version(ABCParse.ABCParse):
    def __init__(self, path = None, *args, **kwargs):

        self.__parse__(locals())

    @property
    def _CONTENTS(self):
        return list(version_path.glob("*"))

    @property
    def hparams(self):
        hparams_path = self._path.joinpath("hparams.yaml")
        if hparams_path.exists():
            return HyperParams(hparams_path)

    @property
    def metrics_df(self):
        metrics_path = self._path.joinpath("metrics.csv")
        if metrics_path.exists():
            return pd.read_csv(metrics_path)

    @property
    def _CKPT_PATHS(self):
        _ckpt_paths = list(self._path.joinpath("checkpoints").glob("*"))
        return [pathlib.Path(path) for path in _ckpt_paths]

    @property
    def _SORTED_CKPT_KEYS(self):
        epochs = list(self.ckpts.keys())
        _epochs = sorted([epoch for epoch in epochs if epoch != "last"])
        if "last" in epochs:
            _epochs.append("last")
        return _epochs

    @property
    def ckpts(self):
        if not hasattr(self, "_CHECKPOINTS"):
            self._CHECKPOINTS = {}
            for ckpt_path in self._CKPT_PATHS:
                ckpt = Checkpoint(ckpt_path)
                self._CHECKPOINTS[ckpt.epoch] = ckpt
        return self._CHECKPOINTS


class ModelLoader(ABCParse.ABCParse):
    def __init__(self, project_path = None, version = None, *args, **kwargs):
        self.__parse__(locals())
        self._INFO = utils.InfoMessage()
        self._validate_version()

    @property
    def project(self):
        if not hasattr(self, "_project"):
            self._project = Project(path=self._project_path)
        return self._project

    @property
    def _VERSION_KEY(self):

        if not hasattr(self, "_VERSION_KEY_PRIVATE"):
            if self._version is None:
                version_key = max(self.project._VERSION_PATHS)
                self._INFO(f"Version not provided. Defaulting to: '{version_key}'")
                self._VERSION_KEY_PRIVATE = version_key
            else:
                self._VERSION_KEY_PRIVATE = f"version_{self._version}"

        return self._VERSION_KEY_PRIVATE

    @property
    def _VERSION_PATH(self):
        return self.project._VERSION_PATHS[self._VERSION_KEY]

    @property
    def version(self):
        if not hasattr(self, "_VERSION"):
            self._VERSION = Version(path=self._VERSION_PATH)
        return self._VERSION

    @property
    def hparams(self):
        return self.version.hparams()

    @property
    def _MODEL_TYPE(self):
        return self.hparams["name"].split(":")[0].replace("-", "_")

    @property
    def model(self):
        if not hasattr(self, "_model"):
            LitModel = getattr(lightning_models, self._MODEL_TYPE)
            self._model = LitModel(**self.hparams)
        return self._model

    @property
    def ckpt(self):
        return self.version.ckpts[self.epoch]
    
    @property
    def epoch(self):
        if not hasattr(self, "_epoch"):
            self._epoch = "last"
            self._INFO(f"Epoch not provided. Defaulting to: '{self._epoch}'")
        return self._epoch

    def _validate_epoch(self):
            
        msg = f"ckpt at epoch: {self.epoch} does not exist. Choose from: {self.version._SORTED_CKPT_KEYS}"
        assert self.epoch in self.version.ckpts, msg

    def _validate_version(self):
        
        version_key = self._VERSION_KEY
        available_versions = [
            attr for attr in self.project.__dir__() if attr.startswith("version_")
        ]

        msg = f"{version_key} not found in project. Available versions: {available_versions}"
        assert version_key in self.project._VERSION_PATHS, msg

    def __call__(
        self,
        epoch: Optional[Union[str, int]] = None,
        plot_state_change: Optional[bool] = False,
        *args,
        **kwargs,
    ) -> lightning.LightningModule:
        
        """
        Parameters
        ----------
        epoch: Optional[Union[str, int]], default = None
        
        plot_state_change: Optional[bool], default = False
        
        Returns
        -------
        model: 
        """
        
        self.__update__(locals())

        self._validate_epoch()

        model = self.model

        if plot_state_change:
            torch_nets.pl.weights_and_biases(model.state_dict())
            
        ckpt_path = self.ckpt.path
        self._INFO(f"Loading model from ckpt: \n\t'{ckpt_path}'")
        model = model.load_from_checkpoint(ckpt_path)
        
        if plot_state_change:
            torch_nets.pl.weights_and_biases(model.state_dict())

        return model
    
def load_diffeq(
    project_path: Union[pathlib.Path, str],
    version: Optional[int] = None,
    epoch: Optional[Union[int, str]] = None,
) -> lightning.LightningModule:

    """
    Load DiffEq from project_path, version [optional], and epoch [optional].
    
    Parameters
    ----------
    project_path: Union[pathlib.Path, str]

    version: Optional[int], default = None

    epoch: Optional[Union[int, str]], default = None

    Returns
    -------
    DiffEq: lightning.LightningModule
    """

    model_loader = ModelLoader(project_path=project_path, version=version)
    return model_loader(epoch=epoch)

def load_model(
    adata: anndata.AnnData,
    project_path: Union[pathlib.Path, str],
    version: Optional[int] = None,
    epoch: Optional[Union[int, str]] = None,
):

    diffeq = load_diffeq(
        project_path=project_path,
        version=version,
        epoch=epoch,
    )
    model = scDiffEq(**dict(diffeq.hparams))
    model.configure_data(adata)
    model.configure_kNN()
    model.configure_model(diffeq)

    return model




from ..core import utils

# from ._x_use import fetch_formatted_data

import anndata
import adata_query
import ABCParse
import anndata as a
import pandas as pd
import numpy as np
import annoy

from typing import Union, List, Dict
from typing import List, Optional
# NoneType = type(None)
# class NeighborCounter(utils.ABCParse):
#     def __init__(self, adata, X_neighbors, max_only: bool = False):

#         self.__parse__(locals(), public=[None])

#     @property
#     def _N_NEIGHBORS(self):
#         return self._X_neighbors.shape[1]

#     @property
#     def _NN_ADATA(self):
#         if not hasattr(self, "_nn_adata"):
#             self._nn_adata = self._adata[self._X_neighbors.flatten()]
#         return self._nn_adata

#     def forward(self, obs_key):
#         labelled_neighbor_df = pd.DataFrame(
#             self._NN_ADATA.obs[obs_key].to_numpy().reshape(-1, self._N_NEIGHBORS).T
#         )
#         return self._count_values(labelled_neighbor_df)

#     def _to_list(self, obs_keys):
#         if isinstance(obs_keys, str):
#             obs_keys = [obs_keys]
#         return obs_keys

#     def _count_values_enumerate(self, col: pd.Series) -> dict:
#         return col.value_counts().to_dict()

#     def _count_values_max(self, col: pd.Series) -> str:
#         return col.value_counts().idxmax()

#     def _count_values(self, labelled_neighbor_df: pd.DataFrame) -> List[Dict]:
#         """
#         labelled_neighbor_df
#         """
#         if not self._max_only:
#             func = self._count_values_enumerate
#         else:
#             func = self._count_values_max

#         counts = [
#             func(labelled_neighbor_df[col]) for col in labelled_neighbor_df.columns
#         ]

#         if not self._max_only:
#             return pd.DataFrame(counts).fillna(0)
#         return counts

#     def __call__(
#         self, obs_keys: Union[List[str], str]
#     ) -> Union[Dict[str, pd.DataFrame], pd.DataFrame]:

#         """"""

#         obs_keys = self._to_list(obs_keys)

#         Counts = {}
#         for obs_key in obs_keys:
#             Counts[obs_key] = self.forward(obs_key)

#         if self._max_only:
#             return pd.DataFrame(Counts)
#         return Counts

class kNN(ABCParse.ABCParse):
    _IDX_BUILT = False

    def __init__(
        self,
        adata: anndata.AnnData,
        use_key: str = "X_pca",
        metric: str = "euclidean",
        n_neighbors: int = 20,
        n_trees: int = 10,
        *args,
        **kwargs,
    ):
        self.__parse__(locals(), public=[None])

#     @property
#     def X_basis(self):
#         """Basis data on which the graph is built."""
#         if not hasattr(self, "_X_basis"):
#             self._X_basis = fetch_formatted_data(
#                 self._adata, use_key=self._use_key, torch=False
#             )

#         return self._X_basis

    def X_use(self):
        if not hasattr(self, "_X_use"):
            self._X_use = adata_query.fetch(self._adata, key = self._use_key, torch = False)            
        return self._X_use

    @property
    def _N_CELLS(self) -> int:
        return self.X_basis.shape[0]

    @property
    def _N_DIMS(self) -> int:
        return self.X_basis.shape[1]

    def _build_annoy_index(self) -> annoy.AnnoyIndex:

        idx = annoy.AnnoyIndex(self._N_DIMS, self._metric)
        [idx.add_item(i, self.X_basis[i]) for i in range(self._N_CELLS)]
        idx.build(self._n_trees)

        return idx

    @property
    def idx(self) -> annoy.AnnoyIndex:

        """"""

        if not hasattr(self, "_idx"):
            self._idx = self._build_annoy_index()
        return self._idx

    def query(self, X_query: np.ndarray) -> np.ndarray:
        """
        Key function to search for neighbor cells of given observations.

        Parameters:
        -----------
        X_query: np.ndarray
            Observations of size: [cells x n_dims].

        Returns:
        --------
        np.ndarray
            Query results composed neighbors.
        """

        return np.array(
            [
                self.idx.get_nns_by_vector(X_query[i], self._n_neighbors)
                for i in range(X_query.shape[0])
            ]
        )

    def count(
        self,
        X_neighbors: np.ndarray,
        obs_keys: Union[List[str], str],
        max_only: bool = False,
    ) -> Union[Dict[str, pd.DataFrame], pd.DataFrame]:

        """
        Count neighbor labels. If `max_only` is True. Provide the predicted label rather than value counts.

        Parameters
        ----------
        X_neighbors: np.ndarray

        obs_keys: Union[List[str], str]

        max_only: bool

        Returns
        -------

        """
        ...

#         neighbor_counter = NeighborCounter(self._adata, X_neighbors, max_only=max_only)
#         return neighbor_counter(obs_keys)

#     def _aggregate(
#         self, X_query, obs_keys: Union[List[str], str], max_only: bool = False
#     ) -> pd.DataFrame:
#         """Queries and counts"""

#         X_nn = self.query(X_query)
#         return self.count(X_nn, obs_keys=obs_keys, max_only=max_only)

#     def _multi_aggregate(
#         self, X_query, obs_keys: str, max_only: bool = False
#     ) -> List[pd.DataFrame]:

#         """
#         Queries and counts across a stacked (3-D) X_query input.
#         """
#         _df_list = [
#             self.aggregate(X_query=X_, obs_keys=obs_keys, max_only=max_only)
#             for X_ in X_query

    def count(self, query_result, obs_key: str, max_only: bool = False, n_neighbors: Optional[int] = None):
        
        if n_neighbors is None:
            n_neighbors = self._n_neighbors
        
        nn_adata = self._adata[query_result.flatten()]
        query_df = pd.DataFrame(
            nn_adata.obs[obs_key].to_numpy().reshape(-1, n_neighbors).T
        )
        del nn_adata

        if not max_only:
            return [
                self._count_values(query_df[i]) for i in query_df.columns
            ]  # list of dicts
        return [
            self._max_count(query_df[i]) for i in query_df.columns
        ]  # list of values

    def aggregate(self, X_query, obs_key: str, max_only: bool = False, n_neighbors: Optional[int] = None):
        _df = (
            pd.DataFrame(
                self.count(
                    query_result=self.query(X_query=X_query),
                    obs_key=obs_key,
                    max_only=max_only,
                    n_neighbors = n_neighbors,
                )
            )
            .fillna(0)
            .sort_index(axis=1)
        )
        if not max_only:
            return _df
        return _df.rename({0:obs_key}, axis = 1)
            

    def multi_aggregate(
        self, X_query, obs_key: str, max_only: bool = False, n_neighbors: Optional[int] = None,
    ) -> List[pd.DataFrame]:
        _list_of_dfs = [
            self.aggregate(
                X_query=X_query[i],
                obs_key=obs_key,
                max_only=max_only,
                n_neighbors = n_neighbors,
            )
            for i in range(len(X_query))
        ]
        return _df_list

    def aggregate(
        self, X_query, obs_keys: str, max_only: bool = False
    ) -> Union[List[pd.DataFrame], pd.DataFrame]:
        """"""

        KWARGS = {"X_query": X_query, "obs_keys": obs_keys, "max_only": max_only}

        if X_query.ndim == 3:
            return self._multi_aggregate(**KWARGS)
        if X_query.ndim == 2:
            return self._aggregate(**KWARGS)
        raise ValueError(f"X_query.ndim should be 2 or 3. Received: {X_query.ndim}")

    def __repr__(self):
        return f"kNN: {self._use_key}"




__module_name__ = "__init__.py"
__doc__ = """Top-level __init__ for the scDiffEq package."""
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu"])


import os as _os


from ._version import _PackageVersion

_package_version = _PackageVersion()
__version__ = __VERSION__ = __Version__ = _package_version()

_os.environ["KEOPS_VERBOSE"] = "0"


# -- import model API: -------------------------------------------------------------------
from .core._scdiffeq import scDiffEq


# -- import sub-packages: ----------------------------------------------------------------
from . import core
from . import io
# from . import plotting as pl
from . import tools as tl
from . import datasets

from . import _backend_utilities as utils





# -- import packages: ----------------------------------------------------------
import os
import abc
import numpy as np
import anndata
import ABCParse

# --
# to avoid requirement of scvelo
# import scvelo as scv
import pip

def import_or_install(package: str):
    try:
        __import__(package)
    except ImportError:
        pip.main(['install', package])  
        __import__(package)
# --

# -- import local dependencies: ------------------------------------------------
from ..core import utils


# -- set typing: ---------------------------------------------------------------
from typing import Union
NoneType = type(None)


# -- Helper classes: -----------------------------------------------------------
class AnnDataAttributeInspector(ABCParse.ABCParse):
    def __init__(self):
        ...

    @property
    @abc.abstractmethod
    def _KEY_SET(self):
        ...

    def transfer(self, old_adata: anndata.AnnData, new_adata: anndata.AnnData):

        for group, key_set in self._KEY_SET.items():
            for key in key_set:
                getattr(new_adata, group)[key] = getattr(old_adata, group)[key]


class NearestNeighborAnnDataInspector(AnnDataAttributeInspector):
    def __init__(
        self,
        obsp_neighbor_keys=["distances", "connectivities"],
        uns_neighbor_keys=["neighbors"],
    ):
        super()
        self.__parse__(locals(), public=[None])

    @property
    def _KEY_SET(self):
        return {"obsp": self._obsp_neighbor_keys, "uns": self._uns_neighbor_keys}

    @property
    def has_obsp_neighbor_keys(self):
        return np.all([key in self.adata.obsp for key in self._obsp_neighbor_keys])

    @property
    def has_uns_neighbor_keys(self):
        return np.all([key in self.adata.uns_keys() for key in self._uns_neighbor_keys])

    @property
    def has_nn(self):
        return np.all([self.has_obsp_neighbor_keys, self.has_uns_neighbor_keys])

    def __call__(self, adata):
        self.__update__(locals())
        return self.has_nn


class VelocityGraphInspector(AnnDataAttributeInspector):
    def __init__(self, velo_key="Z_drift", basis="umap"):
        super()
        self.__parse__(locals(), public=[None])

    @property
    def _KEY_SET(self):
        """_SCVELO_ADDED_GRAPH_KEYS"""
        return {
            "uns": [
                f"{self._velo_key}_graph",
                f"{self._velo_key}_graph_neg",
                f"{self._velo_key}_params",
            ],
            #             "obsm": [f"{self._velo_key}_{self._basis}"],
            "obs": [f"{self._velo_key}_self_transition"],
        }

    @property
    def _VELO_GRAPH_COMPUTED(self):

        _GROUPS = []
        for group, key_set in self._KEY_SET.items():
            _GROUPS.append(
                np.all([key in getattr(self.adata, group) for key in key_set])
            )
        return np.all(_GROUPS)

    def __call__(self, adata):
        self.__update__(locals())
        return self._VELO_GRAPH_COMPUTED
    
class VelocityAnnDataFormatter(ABCParse.ABCParse):
    """Reformatted, minimal adata for velo plots"""

    def __init__(
        self,
        state_key="X_pca",
        drift_layer_key="Z_drift",
        diffusion_layer_key="Z_diffusion",
        drift_obs_key="drift",
        diffusion_obs_key="diffusion",
    ):
        self.__parse__(locals(), public=[None])

        self._NN_INSPECTOR = NearestNeighborAnnDataInspector()
        self._VELO_GRAPH_INSPECTOR = VelocityGraphInspector()

    @property
    def Z_state(self):
        return self.adata.obsm[self._state_key]

    @property
    def X_umap(self):
        return self.adata.obsm["X_umap"]

    @property
    def Z_drift(self):
        return self.adata.obsm[self._drift_layer_key]

    @property
    def Z_diffusion(self):
        return self.adata.obsm[self._diffusion_layer_key]

    @property
    def obs_df(self):
        obs_df = self.adata.obs[[self._drift_obs_key, self._diffusion_obs_key]].copy()
        for col in obs_df.columns:
            obs_df[col] = obs_df[col].astype(float).values
        return obs_df

    @property
    def HAS_NN(self):
        return self._NN_INSPECTOR(self.adata)

    @property
    def HAS_VELOCITY_GRAPH(self):
        return self._VELO_GRAPH_INSPECTOR(self.adata)

    @property
    def adata_velo(self):
        if not hasattr(self, "_adata_velo"):
            self._adata_velo = anndata.AnnData(
                X=self.Z_state,
                dtype=self.Z_state.dtype,
                layers={
                    self._drift_layer_key: self.Z_drift,
                    self._diffusion_layer_key: self.Z_diffusion,
                    "spliced": self.Z_state,
                },
                obs=self.obs_df,
                obsm={"X_umap": self.X_umap},
            )
            if self._NN_INSPECTOR(self.adata):
                self._NN_INSPECTOR.transfer(self.adata, self._adata_velo)
            else:
                scv.pp.neighbors(self._adata_velo, use_rep="X")
        return self._adata_velo

    def __call__(self, adata: anndata.AnnData):

        self.__update__(locals())
        return self.adata_velo


# -- Main operational class: ---------------------------------------------------
class VelocityStreamPlot(ABCParse.ABCParse):
    _GRAPH_COMPUTED = []

    def __init__(
        self,
        state_key="X_pca",
        drift_layer_key="Z_drift",
        diffusion_layer_key="Z_diffusion",
        drift_obs_key="drift",
        diffusion_obs_key="diffusion",
    ):
        self.__parse__(locals(), public=["adata"])

        self.velocity_formatter = VelocityAnnDataFormatter(
            **utils.function_kwargs(VelocityAnnDataFormatter, self._PARAMS)
        )

    @property
    def adata_velo(self):
        if not hasattr(self, "_adata_velo"):
            self._adata_velo = self.velocity_formatter(self.adata)
        return self._adata_velo

    def _compute_velocity_graph(
        self, velocity_key="Z_drift", n_jobs=int(os.cpu_count() / 2), *args, **kwargs
    ):
        self._velo_key = velocity_key
        if self.velocity_formatter.HAS_VELOCITY_GRAPH:
            self.velocity_formatter._VELO_GRAPH_INSPECTOR.transfer(
                self.adata,
                self.adata_velo,
            )
        else:
            scv.tl.velocity_graph(
                self.adata_velo, vkey=velocity_key, n_jobs=n_jobs, **kwargs
            )
            self.velocity_formatter._VELO_GRAPH_INSPECTOR.transfer(
                self.adata_velo, self.adata
            )

        self._GRAPH_COMPUTED.append(velocity_key)

    @property
    def _VMIN(self):
        return np.quantile(self.adata_velo.obs[self.color], self.color_quantile)

    @property
    def _VMAX(self):
        return np.quantile(self.adata_velo.obs[self.color], (1 - self.color_quantile))

    def __call__(
        self,
        adata,
        velocity_key="Z_drift",
        color="diffusion",
        basis="umap",
        color_quantile=0.05,
        cmap="plasma",
        save: Union[bool, str] = False,
        **kwargs,
    ):

        self.__update__(locals())

        self._adata_velo = self.velocity_formatter(self.adata)

        if not velocity_key in self._GRAPH_COMPUTED:
            self._compute_velocity_graph(velocity_key=self.velocity_key)

        scv.pl.velocity_embedding_stream(
            self.adata_velo,
            vkey=self.velocity_key,
            basis=self.basis,
            color=self.color,
            cmap=self.cmap,
            vmin=self._VMIN,
            vmax=self._VMAX,
            save=self.save,
            **kwargs,
        )
        emb_key_added = f"{velocity_key}_{basis}"
        adata.obsm[emb_key_added] = self.adata_velo.obsm[emb_key_added]
        

# -- API-facing function: ------------------------------------------------------
def velocity_stream(
    adata: anndata.AnnData,
    velocity_key: str = "Z_drift",
    color: str = "diffusion",
    state_key: str = "X_pca",
    drift_layer_key: str = "Z_drift",
    diffusion_layer_key: str = "Z_diffusion",
    drift_obs_key: str = "drift",
    diffusion_obs_key: str = "diffusion",
    basis: str = "umap",
    color_quantile: float = 0.05,
    cmap: str ="plasma",
    save: bool = False,
    **kwargs,
):
    """
    Parameters:
    -----------
    adata
        type: anndata.AnnData
        default: [ REQUIRED ]
        
    velocity_key
        type: str
        default: "Z_drift"
        
    color
        type: str
        default: "diffusion"
        
    state_key
        type: str
        default: "X_pca"
        
    drift_layer_key
        type: str
        default: "Z_drift"
        
    diffusion_layer_key
        type: str
        default: "Z_diffusion"
        
    drift_obs_key
        type: str
        default: "drift"
        
    diffusion_obs_key
        type: str
        default: "diffusion"
        
    basis
        type: str
        default: "umap"
        
    color_quantile
        type: float
        default: 0.05
        
    cmap
        type: str
        default: "plasma"
        
    save
        type: bool
        default: False
    
    Returns:
    --------
    None, plots velocity stream plot.
    
    Notes:
    ------
    1.  Increased flexibility with passed parameters is a TODO, will
        be mostly through **kwargs.
    """
    
    
    import_or_install("scvelo")
    
    velo_stream_plot = VelocityStreamPlot(
        state_key=state_key,
        drift_layer_key=drift_layer_key,
        diffusion_layer_key=diffusion_layer_key,
        drift_obs_key=drift_obs_key,
        diffusion_obs_key=diffusion_obs_key,
    )
    velo_stream_plot(
        adata=adata,
        velocity_key=velocity_key,
        color=color,
        basis=basis,
        color_quantile=color_quantile,
        cmap=cmap,
        save=save,
    )
    


from IPython import display
import os


def display_img(filename):
    display.display(display.Image(filename=filename))
    
def display_tracked_loss(DiffEqLogger):
    fname = os.path.join(
        DiffEqLogger.versioned_model_outdir, "scDiffEq_fit_loss_tracking.png"
    )
    display_img(filename=fname)



from abc import abstractmethod

class PreTrainMixIn(object):
    """
    Requires user to:
    1. pass pretrain_epochs to the model __init__
    2. def `pretrain_step` method on the model
    """

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.automatic_optimization = False

    def training_step(self, batch, batch_idx, *args, **kwargs):
        if self.PRETRAIN:
            return self.pretrain_step(batch, batch_idx, stage="pretrain")
        return self.step(batch, batch_idx, stage="training")

    def validation_step(self, batch, batch_idx, *args, **kwargs):
        if not self.PRETRAIN:
            return self.step(batch, batch_idx, stage="validation")

    @property
    def PRETRAIN(self):
        return self.COMPLETED_EPOCHS < self.hparams["pretrain_epochs"]
#         return self.current_epoch < self.hparams["pretrain_epochs"]

    @abstractmethod
    def pretrain_step(self):
        ...



import anndata

def idx_to_int_str(adata: anndata.AnnData, drop: bool = False) -> anndata.AnnData:
    """
    """
    adata.obs = adata.obs.reset_index(drop=drop)
    adata.obs.index = adata.obs.index.astype(str)
    return adata